from pink_engine.tiled_game import TiledMapGame, TiledMapGameObject, TiledMapSpriteCollection, \
    TiledMapGameObjectSpriteCollection, TiledMapGameLayer, TiledMapGameTile, TiledMapCamera, \
    TiledMapGameTileAnimated, LAYER_PRIORITY
from pink_engine.orthogonal_tiled_map_commands import ControlCommand
from pink_engine.orthogonal_tiled_map_parallel_processes import *  # noqa
from pink_engine.tileset import TilesetTile, TilesetAnimatedTile
import renpy  # noqa
import pygame  # noqa
import json
import random
import os
from pink_engine.commons import Coord
from math import sqrt, ceil, floor

DEFAULT_MOVEMENT_SPEED = renpy.store.pink_otm_default_npc_move_speed


class OrthogonalTiledMapGameObjectBase(object):
    properties = {}
    width = 0
    height = 0

    def __init__(self):
        """
        :return: The parent class for tiles and objects on an OTM map.
        """
        pass

    @property
    def event_on_activate(self):
        """
        :return: The label of the event that should trigger when interacting with this object.
        :rtype: str|None
        """
        if 'event_on_activate' in self.properties:
            return self.properties['event_on_activate']
        else:
            return None

    @property
    def event_on_activate_args(self):
        """
        :return: the list of args to pass along to the event_on_activate label
        :rtype: list
        """
        if 'event_on_activate_args' in self.properties:
            return eval(self.properties['event_on_activate_args'])
        else:
            return []

    @property
    def event_on_activate_kwargs(self):
        """
        :return: the dict of kwargs to pass along to the event_on_activate label
        :rtype: dict
        """
        if 'event_on_activate_kwargs' in self.properties:
            return eval(self.properties['event_on_activate_kwargs'])
        else:
            return {}

    @property
    def event_on_touch(self):
        """
        :return: The label of the event that should trigger when touching this object.
        :rtype: str|None
        """
        if 'event_on_touch' in self.properties:
            return self.properties['event_on_touch']
        else:
            return None

    @property
    def event_on_touch_args(self):
        """
        :return: the list of args to pass along to the event_on_activate label
        :rtype: list
        """
        if 'event_on_touch_args' in self.properties:
            return eval(self.properties['event_on_touch_args'])
        else:
            return []

    @property
    def event_on_touch_kwargs(self):
        """
        :return: the dict of kwargs to pass along to the event_on_activate label
        :rtype: dict
        """
        if 'event_on_touch_kwargs' in self.properties:
            return eval(self.properties['event_on_touch_kwargs'])
        else:
            return {}

    @property
    def code_on_activate(self):
        """
        :return: The code (in string form) that should run when interacting with this object.
        :rtype: str|None
        """
        if 'code_on_activate' in self.properties:
            return self.properties['code_on_activate']
        else:
            return None

    @property
    def code_on_touch(self):
        """
        :return: The code (in string form) that should run when touching this object.
        :rtype: str|None
        """
        if 'code_on_touch' in self.properties:
            return self.properties['code_on_touch']
        else:
            return None

    @property
    def event_conditional(self):
        """
        :return: The conditions under which this event should trigger. Similar to conditional, but run at the time of
        checking for the presence of events, rather than when rendering the map. Intended to be used for things like
        player orientation.
        :rtype: str
        """
        if "event_conditional" in self.properties:
            return self.properties["event_conditional"]
        else:
            return "True"

    @property
    def event_conditional_met(self):
        """
        :return: Whether or not the event conditional is currently met
        :rtype: bool
        """
        return eval(self.event_conditional)

    @property
    def code_conditional(self):
        """
        :return: The conditions under which this code should trigger. Similar to conditional, but run at the time of
        checking for the presence of code, rather than when rendering the map. Intended to be used for things like
        player orientation.
        :rtype: str
        """
        if "code_conditional" in self.properties:
            return self.properties["code_conditional"]
        else:
            return "True"

    @property
    def code_conditional_met(self):
        """
        :return: Whether or not the code conditional is currently met
        :rtype: bool
        """
        return eval(self.code_conditional)

    @property
    def base_offset_x_start(self):
        """
        :return: The x offset of the start of the object's base (where the object begins for mapping and event
        purposes in relation to where the image of the object begins)
        :rtype: int
        """
        if "base_offset_x_start" in self.properties:
            return self.properties["base_offset_x_start"]
        else:
            return 0

    @property
    def base_offset_x_end(self):
        """
        :return: The x offset of the end of the object's base (where the object ends for mapping and event
        purposes in relation to where the image of the object begins)
        :rtype: int
        """
        if "base_offset_x_end" in self.properties:
            return self.properties["base_offset_x_end"]
        else:
            return self.width

    @property
    def base_offset_y_start(self):
        """
        :return: The y offset of the start of the object's base (where the object begins for mapping and event
        purposes in relation to where the image of the object begins)
        :rtype: int
        """
        if "base_offset_y_start" in self.properties:
            return self.properties["base_offset_y_start"]
        else:
            return 0

    @property
    def base_offset_y_end(self):
        """
        :return: The y offset of the end of the object's base (where the object ends for mapping and event
        purposes in relation to where the image of the object begins)
        :rtype: int
        """
        if "base_offset_y_end" in self.properties:
            return self.properties["base_offset_y_end"]
        else:
            return self.height

    @property
    def move_from(self):
        """
        :return: The movement allowed from a square occupied by this object.
        :rtype: str|None
        """
        if "move_from" in self.properties:
            return self.properties["move_from"]
        else:
            return None

    @property
    def move_to(self):
        """
        :return: The movement allowed to a square occupied by this object.
        :rtype: str|None
        """
        if "move_to" in self.properties:
            return self.properties["move_to"]
        else:
            return None


class OrthogonalTiledMapGameTileAnimated(TiledMapGameTileAnimated, OrthogonalTiledMapGameObjectBase):
    def __init__(self, parent_map, parent_tile):
        """
        An animated tile in an orthogonal tiled map game.
        :param OrthogonalTiledMap parent_map: The map object that this object is displayed on.
        :param TilesetAnimatedTile parent_tile: The tileset object this tile is inheriting properties from.
        """
        TiledMapGameTileAnimated.__init__(self, parent_map, parent_tile)
        OrthogonalTiledMapGameObjectBase.__init__(self)


class OrthogonalTiledMapGameTile(TiledMapGameTile, OrthogonalTiledMapGameObjectBase):
    def __init__(self, parent_map, parent_tile):
        """
        An non-animated tile in an orthogonal tiled map game.
        :param OrthogonalTiledMap parent_map: The map object that this object is displayed on.
        :param TilesetTile parent_tile: The tileset object this tile is inheriting properties from.
        """
        TiledMapGameTile.__init__(self, parent_map, parent_tile)
        OrthogonalTiledMapGameObjectBase.__init__(self)


class OrthogonalTiledMapGameObject(TiledMapGameObject, OrthogonalTiledMapGameObjectBase):
    def __init__(self, parent_rpg_tile, x, y, width, height, layer, properties):
        """
        A map object in an orthogonal tiled map game.
        :param TilesetTile|TiledMapSpriteCollection parent_rpg_tile: The tileset object this tile is inheriting
        properties from.
        :param int x: Distance in pixels between the object's left side and the left border of the map.
        :param int y: Distance in pixels between the object's top side and the top border of the map.
        :param int width: Width of the object's image, in pixels.
        :param int height: Height of the object's image in pixels.
        :param TiledMapGameLayer layer: The layer this object should be placed on.
        :param dict properties: Dictionary of pink engine properties that applies to this object.
        """
        TiledMapGameObject.__init__(self, parent_rpg_tile, x, y, width, height, layer, properties)
        OrthogonalTiledMapGameObjectBase.__init__(self)

        if type(self) is OrthogonalTiledMapGameObject and self.ref_name is not None:
            # Type comparison ensures this only happens at the end of the final loading
            setattr(renpy.store, self.ref_name, self)

    def pixel_to_centralize_on_coord(self, x_coord, y_coord):
        """
        Returns the x and y this object should be set to in order for its base to be centered on the given coordinates.
        :param int x_coord: The given x coordinate
        :param int y_coord: the given y coordinate
        """
        tile_width = self.parent.tile_size.x
        tile_height = self.parent.tile_size.y

        x_pixel = (x_coord * tile_width) + ceil(float(tile_width - self.base_width) / 2.0) - self.base_offset_x_start
        y_pixel = (y_coord * tile_height) + ceil(float(tile_height - self.base_height) / 2.0) - self.base_offset_y_start

        return x_pixel, y_pixel

    @property
    def base_offset_x_start(self):
        """
        :return: The x offset of the start of the object's base (where the object begins for mapping and event
        purposes in relation to where the image of the object begins)
        :rtype: int
        """
        if "base_offset_x_start" in self.properties:
            return int(self.properties["base_offset_x_start"] * self.xzoom)
        else:
            return 0

    @property
    def base_offset_x_end(self):
        """
        :return: The x offset of the end of the object's base (where the object ends for mapping and event
        purposes in relation to where the image of the object begins)
        :rtype: int
        """
        if "base_offset_x_end" in self.properties:
            return int(self.properties["base_offset_x_end"] * self.xzoom)
        else:
            return self.width

    @property
    def base_offset_y_start(self):
        """
        :return: The y offset of the start of the object's base (where the object begins for mapping and event
        purposes in relation to where the image of the object begins)
        :rtype: int
        """
        if "base_offset_y_start" in self.properties:
            return int(self.properties["base_offset_y_start"] * self.yzoom)
        else:
            return 0

    @property
    def base_offset_y_end(self):
        """
        :return: The y offset of the end of the object's base (where the object ends for mapping and event
        purposes in relation to where the image of the object begins)
        :rtype: int
        """
        if "base_offset_y_end" in self.properties:
            return int(self.properties["base_offset_y_end"] * self.yzoom)
        else:
            return self.height

    @property
    def camera_offset_x(self):
        """
        :return: By how many pixels to offset the camera on the horizontal axis when focused on this object.
        :rtype: int
        """
        if 'camera_offset_x' in self.properties:
            return self.properties['camera_offset_x']
        else:
            return 0

    @camera_offset_x.setter
    def camera_offset_x(self, value):
        """
        :param int value: New camera offset x value
        """
        self.properties['camera_offset_x'] = value

    @property
    def camera_offset_y(self):
        """
        :return: By how many pixels to offset the camera on the vertical axis when focused on this object.
        :rtype: int
        """
        if 'camera_offset_y' in self.properties:
            return self.properties['camera_offset_y']
        else:
            return 0

    @camera_offset_y.setter
    def camera_offset_y(self, value):
        """
        :param int value: New camera offset x value
        """
        self.properties['camera_offset_y'] = value

    @property
    def coords(self):
        """
        :return: The list of coordinates covered by this object's base.
        :rtype: list
        """
        return self.parent.base_grid.map_elements[self.displayable_id].coords

    @property
    def central_coord(self):
        """
        :return: The coordinate covered by the center of this object's base.
        :rtype: Coord
        """
        return self.parent.base_grid.map_elements[self.displayable_id].central_coord

    @property
    def base_x_start(self):
        """
        :return: The distance in pixels between the left side of the base of this object and the left border of the
        map.
        :rtype: int
        """
        return self.parent.base_grid.map_elements[self.displayable_id].base_x_start

    @property
    def base_x_end(self):
        """
        :return: The distance in pixels between the right side of the base of this object and the left border of the
        map.
        :rtype: int
        """
        return self.parent.base_grid.map_elements[self.displayable_id].base_x_end

    @property
    def base_y_start(self):
        """
        :return: The distance in pixels between the upper side of the base of this object and the upper border of the
        map.
        :rtype: int
        """
        return self.parent.base_grid.map_elements[self.displayable_id].base_y_start

    @property
    def base_y_end(self):
        """
        :return: The distance in pixels between the bottom side of the base of this object and the upper border of the
        map.
        :rtype: int
        """
        return self.parent.base_grid.map_elements[self.displayable_id].base_y_end

    @property
    def base_width(self):
        """
        :return: This object's base's width in pixels.
        :rtype: int
        """
        return self.base_offset_x_end - self.base_offset_x_start

    @property
    def base_height(self):
        """
        :return: This object's base's height in pixels.
        :rtype: int
        """
        return self.base_offset_y_end - self.base_offset_y_start

    @property
    def occultates_as_y(self):
        """
        :return: Which y coordinate this object occultates as.
        :rtype: int
        """
        if 'occultates_as_y' in self.properties:
            return self.properties['occultates_as_y']
        if self.parent.orientation == "orthogonal":
            return int(floor((self.y + self.base_offset_y_end - 1) / self.parent.tile_size.y))
        else:
            return self.y + self.height - 1


class OrthogonalTiledMapGameObjectMobile(OrthogonalTiledMapGameObject):
    movement_commands = {}

    def __init__(self, parent_rpg_tile, x, y, width, height, layer, properties):
        """
        A map object in an orthogonal tiled map game.
        :param TilesetTile|TiledMapSpriteCollection parent_rpg_tile: The tileset object this tile is inheriting
        properties from.
        :param int x: Distance in pixels between the object's left side and the left border of the map.
        :param int y: Distance in pixels between the object's top side and the top border of the map.
        :param int width: Width of the object's image, in pixels.
        :param int height: Height of the object's image in pixels.
        :param TiledMapGameLayer layer: The layer this object should be placed on.
        :param dict properties: Dictionary of pink engine properties that applies to this object.
        """
        OrthogonalTiledMapGameObject.__init__(self, parent_rpg_tile, x, y, width, height, layer, properties)

        self._last_gt = 0.0
        self._command_gt_start = 0.0

        if self.should_load_consistent(properties):
            old_incarnation = getattr(renpy.store, properties['ref_name'],
                                      None)  # type: OrthogonalTiledMapGameObjectMobile
            self.control_stack = old_incarnation.control_stack

            # Command handling variables
            self._current_command = old_incarnation._current_command  # type: ControlCommand
            self._command_gt_end = old_incarnation._command_gt_end - old_incarnation._last_gt
            self._command_x_start = old_incarnation.x  # type: int
            self._command_y_start = old_incarnation.y  # type: int
            self.x = old_incarnation.x  # type: int
            self.y = old_incarnation.y  # type: int
            self._command_x_end = old_incarnation._command_x_end
            self._command_y_end = old_incarnation._command_y_end
            self.prev_coord = old_incarnation.prev_coord

        else:
            self.control_stack = self.default_control_stack

            # Command handling variables
            self._current_command = None
            self._command_gt_end = 0.0
            self._command_x_start = self.x
            self._command_y_start = self.y
            self._command_x_end = self.x
            self._command_y_end = self.y
            self.prev_coord = None

        # Saved Command handling variables
        self._save_x = self._command_x_end
        self._save_y = self._command_y_end
        self._saved_control_stack = []
        self._saved_repeat_commands = self.repeat_commands

        # Set to True if object should seize all movement.
        self.frozen = False

        if type(self) is OrthogonalTiledMapGameObjectMobile and self.ref_name is not None:
            # Type comparison ensures this only happens at the end of the final loading
            setattr(renpy.store, self.ref_name, self)

    def per_tick(self, gt):
        """
        What this object should do per game tick.
        :param float gt: The amount of seconds of game time that have passed.
        """
        if gt < self._last_gt:
            self.rebase_st(gt)
        if not self.frozen:
            if self._current_command is not None:
                self._update_command(gt)
            if self._current_command is None and len(self.control_stack) > 0:
                self._consume_command(gt)
        self._last_gt = gt

    def set_to_coords(self, x_coord, y_coord, **kwargs):
        """
        Instantly moves this object so that its base is centered on the given coordinates. Note that this overrides
        the saved x and saved y as well, so if used during a dynamic event, it will cause the object to not return
        to its starting location.
        :param int x_coord: The given x coordinate
        :param int y_coord: The given y coordinate.
        """
        x, y = self.pixel_to_centralize_on_coord(x_coord, y_coord)
        self.set_to_x_y(x, y, **kwargs)

    def set_to_x_y(self, x, y, override_prev_coord=True, **kwargs):
        """
        Instantly moves this object so that the left side of its image is x pixels from the left border of the map and
        the top side of its image is y pixels from the top right border of the map. Note that this overrides
        the saved x and saved y as well, so if used during a dynamic event, it will cause the object to not return
        to its starting location.
        :param bool override_prev_coord: Whether to override the previous coordinate
        :param int x: The given x
        :param int y: The given y
        """
        if self.is_conditional:  # So that enabling a conditional object and then immediately moving it is possible.
            self.parent.check_conditional_object(self)

        self.x, self.y = x, y
        self._save_x, self._save_y = x, y
        self._command_x_start = x
        self._command_y_start = y
        self._command_x_end = x
        self._command_y_end = y

        self.parent.move_element(self.displayable_id, x, y)
        self.parent.base_grid.move_element(self.displayable_id, x, y)

        if override_prev_coord:
            self.prev_coord = self.central_coord

    def freeze(self):
        """
        Freezes this object (causing it to remain motionless in place until unfrozen)
        """
        self.frozen = True

    def unfreeze(self, increment_gt=True):
        """
        Unfreezes this object
        :param bool increment_gt: Whether or not to increment the GT. When unfreezing on game load, this should not be
        done. Otherwise, it should be done.
        """
        if increment_gt:
            self.increment_gt(self._last_gt)
        self.frozen = False

    def increment_gt(self, gt_diff):
        """
        Increments the movement start and end gt by the given amount. To be used after freezes.
        :param float gt_diff: The value by which to increment the gt.
        """
        self._command_gt_start += gt_diff
        self._command_gt_end += gt_diff

    def save(self):
        """
        Saves this object's current location and control stack, so it can be returned to using the load_and_continue
        function
        """
        self._save_x = self._command_x_end
        self._save_y = self._command_y_end
        self._saved_control_stack = self.control_stack[:]  # py2
        self._saved_repeat_commands = self.repeat_commands

    def save_and_halt(self):
        """
        Saves this object's current location and control stack, so it can be returned to using the load_and_continue
        function, and then causes it to halt (making it finish its current movement and then stand still until given
        new instructions)
        """
        self.save()
        self.control_stack = []

    def save_and_default_interaction(self):
        """
        Saves this object's current location and control stack, so it can be returned to using the load_and_continue
        function, and then causes it to execute its default interaction, after which it halts
        (making it finish its current movement and then stand still until given new instructions)
        """
        self.save()
        self.default_interaction()

    def default_interaction(self):
        """
        Makes this object execute its default interaction, and then halt (making it finish its current movement and
        then stand still until given new instructions)
        """
        self.control_stack = self.default_interaction_stack

    def load_and_continue(self):
        """
        Restores this object's saved location and control stack, and continues following it.
        """
        self.control_stack = self._saved_control_stack
        self.repeat_commands = self._saved_repeat_commands
        self.set_to_x_y(self._save_x, self._save_y, override_prev_coord=False)

    def _consume_command(self, gt):
        """
        Consumes and executes the top control on the control stack
        :param float gt: The current game time in seconds
        """
        command = self.control_stack[0]
        self.movement_commands[command.type](self, command=command, gt=gt)

        if command.instant:  # TODO hardcode a limit?
            self._consume_command(gt)

    def _update_command(self, gt):
        """
        Updates this object's current x and y location depending on how far it is into the current control command
        :param float gt: The current game time in seconds
        """
        if gt >= self._command_gt_end:  # Command is done
            self._finish_executing_command()
        elif gt == self._command_gt_start:  # Command has just started
            self.x = self._command_x_start
            self.y = self._command_y_start
        elif gt > self._command_gt_start:  # anywhere in-between.
            proportion = (gt - self._command_gt_start) / (self._command_gt_end - self._command_gt_start)
            self.x = self._command_x_start + (self._command_x_end - self._command_x_start) * proportion
            self.y = self._command_y_start + (self._command_y_end - self._command_y_start) * proportion
        self.parent.move_element(self.displayable_id, self.x, self.y)

    def rebase_st(self, gt):
        """
        To be called after gt has been reset for whatever reason, causing this object to rebase its current movement
        off the given new gt.
        :param float gt: The new game time in seconds
        """
        self._command_gt_start = float(gt)
        self._command_gt_end -= self._last_gt
        self._command_x_start = self.x
        self._command_y_start = self.y

    def _finish_executing_command(self):
        """
        Finishes executing the current ongoing control command.
        """
        self.x = self._command_x_end
        self.y = self._command_y_end
        self._command_x_start = self.x
        self._command_x_end = self.x
        self._command_y_start = self.y
        self._command_y_end = self.y

        self._current_command = None

        if (
                self.parent.player_object is not None and
                self.parent.player_object.central_coord in self.coords and
                self._command_gt_start != self._command_gt_end
        ):
            self._command_gt_start = self._command_gt_end
            self.parent.check_touch(self)
        self._command_gt_start = self._command_gt_end

    def _command_move_direction(self, command, gt, target_x, target_y, pop=True):
        """
        Move a single square in a non-diagonal direction
        :param ControlCommand command: The control command which caused this movement
        :param float gt: The current game time in seconds
        :param int target_x: The x this object should have at the end of the given movement.
        :param int target_y: The y this object should have at the end of the given movement.
        :param bool pop: If True, removes the current top object from the control stack (affected by the command's
        pop_invalid_move value).
        """
        self._current_command = command

        # get movement_speed
        if command.movement_time is not None:
            movement_time = command.movement_time
        else:
            movement_time = self.movement_speed

        self._command_gt_start = gt
        self._command_gt_end = gt + movement_time
        self._command_x_start, self._command_y_start = self.x, self.y
        if self.can_always_move or self.parent.base_grid.is_move_permitted(self.displayable_id, target_x, target_y):
            self.prev_coord = self.central_coord
            self.parent.base_grid.move_element(self.displayable_id, target_x, target_y)
            self._command_x_end = target_x
            self._command_y_end = target_y
            if pop:
                self.finish_command(command)
        else:
            self._command_x_end, self._command_y_end = self.x, self.y
            if pop and command.pop_invalid_move:
                self.finish_command(command)

    def command_move_left(self, command, gt, pop=True):
        """
        Move a single square to the left
        :param ControlCommand command: The control command which caused this movement
        :param float gt: The current game time in seconds
        :param bool pop: If True, removes the current top object from the control stack (affected by the command's
        pop_invalid_move value).
        """
        self._command_move_direction(
            command=command, gt=gt, target_x=self.x - self.parent.tile_size.x, target_y=self.y, pop=pop)

    def command_move_right(self, command, gt, pop=True):
        """
        Move a single square to the right
        :param ControlCommand command: The control command which caused this movement
        :param float gt: The current game time in seconds
        :param bool pop: If True, removes the current top object from the control stack (affected by the command's
        pop_invalid_move value).
        """
        self._command_move_direction(
            command=command, gt=gt, target_x=self.x + self.parent.tile_size.x, target_y=self.y, pop=pop)

    def command_move_up(self, command, gt, pop=True):
        """
        Move a single square up
        :param ControlCommand command: The control command which caused this movement
        :param float gt: The current game time in seconds
        :param bool pop: If True, removes the current top object from the control stack (affected by the command's
        pop_invalid_move value).
        """
        self._command_move_direction(
            command=command, gt=gt, target_x=self.x, target_y=self.y - self.parent.tile_size.y, pop=pop)

    def command_move_down(self, command, gt, pop=True):
        """
        Move a single square down
        :param ControlCommand command: The control command which caused this movement
        :param float gt: The current game time in seconds
        :param bool pop: If True, removes the current top object from the control stack (affected by the command's
        pop_invalid_move value).
        """
        self._command_move_direction(
            command=command, gt=gt, target_x=self.x, target_y=self.y + self.parent.tile_size.y, pop=pop)

    def command_move_to(self, command, gt):
        """
        Executes and consumes a go_to command, causing an object to move towards a location. The command isn't
        finished until the object has reached the location
        :param ControlCommand command: The command being executed
        :param float gt: The current game time in seconds
        """
        target = command.target

        central_coord = self.central_coord
        if type(target) in [str, unicode]:  # noqa
            target_object = getattr(renpy.store, target)  # Py2
            if target_object is not None:
                target_coord = target_object.central_coord
                target = (target_coord.x, target_coord.y)
            else:
                self.finish_command(command)

        if central_coord.y < target[1]:
            self.command_move_down(command, gt, pop=False)
        elif central_coord.y > target[1]:
            self.command_move_up(command, gt, pop=False)
        elif central_coord.x < target[0]:
            self.command_move_right(command, gt, pop=False)
        elif central_coord.x > target[0]:
            self.command_move_left(command, gt, pop=False)

        # Updates coords after movement
        central_coord = self.central_coord

        # If single step is disabled, then the command isn't consumed until the object reaches the target.
        if (central_coord.x, central_coord.y) == target or command.single_step:
            self.finish_command(command)

    def command_move_random(self, command, gt):
        """
        Executes a move_random command, causing the object to move in a random direction.
        :param ControlCommand command: The command being executed.
        :param float gt: The current game time in seconds
        """
        directions = ["up", "down", "left", "right"]

        for coord in self.coords:
            if "right" in directions and coord.x >= command.max_x:
                directions.remove("right")
            if "left" in directions and coord.x <= command.min_x:
                directions.remove("left")
            if "down" in directions and coord.y >= command.max_y:
                directions.remove("down")
            if "up" in directions and coord.y <= command.min_y:
                directions.remove("up")

        if len(directions) == 0:
            # If no movement is permitted.
            self.finish_command(command)
        else:
            random_direction = random.choice(directions)

            if random_direction == "up":
                self.command_move_up(command, gt)
            elif random_direction == "down":
                self.command_move_down(command, gt)
            elif random_direction == "left":
                self.command_move_left(command, gt)
            if random_direction == "right":
                self.command_move_right(command, gt)

    def command_delay(self, command, gt):
        """
        Executes a delay command, causing the object to wait a number of seconds until consuming its next command.
        :param ControlCommand command: The command being executed.
        :param float gt: The current game time in seconds
        """
        delay_quantity = command.quantity
        if delay_quantity is None:  # Delays default to 1 second.
            delay_quantity = 1.0

        self._command_gt_start = gt
        self._command_gt_end = gt + delay_quantity
        self._current_command = command

        self.finish_command(command)

    def command_change_movement_speed(self, command, gt):
        """
        Executes and consumes a 'change_movement_speed' command, which changes the movement speed of the object.
        Defaults to a speed of 0.14 if no speed is provided.
        :param ControlCommand command: The command being executed.
        :param float gt: The current game time in seconds
        """
        self._current_command = command
        new_speed = command.quantity
        if new_speed is None:
            new_speed = 0.14

        self.movement_speed = new_speed

        if self is self.parent.player_object:
            for follower in renpy.store.pink_otm_followers:
                follower.movement_speed = new_speed

        self._command_gt_start = gt
        self._command_gt_end = gt

        self.finish_command(command)

    def finish_command(self, command):
        """
        Finishes consuming the object on top of the control stack, popping it from the stack. Depending on this object's
        repeat_commands value and the never_repeat value of the given command, the popped command may be re-appended to
        the end of the control stack.
        :param ControlCommand command: The command being executed.
        """
        self.control_stack.pop(0)
        if self.repeat_commands and not command.never_repeat:
            self.control_stack.append(command)
        else:
            del command

    def in_place(self, x_coord=None, y_coord=None, orientation=None, finished_moving=True):
        """
        Returns True if the character is in the given location.
        :param int x_coord: the given x coordinate. None for wildcard.
        :param int y_coord: the given y coordinate. None for wildcard.
        :param str orientation: The given orientation. None for wildcard.
        :param bool finished_moving: Whether the control stack needs to be empty.
        :return:
        """
        if self.displayable_id is None:  # Not currently on the map for whatever reason
            return False
        if x_coord is not None and self.central_coord.x != x_coord:
            return False
        if y_coord is not None and self.central_coord.y != y_coord:
            return False
        if finished_moving and (self.control_stack != [] or self._current_command is not None):
            return False
        return True

    @property
    def repeat_commands(self):
        """
        Returns a boolean that indicates whether this object repeats its command list.
        :return: a boolean that indicates whether this object repeats its command list.
        :rtype: bool
        """
        if 'repeat_commands' in self.properties:
            return self.properties['repeat_commands']
        else:
            return True

    @repeat_commands.setter
    def repeat_commands(self, value):
        """
        :param bool value: The given value to set the repeat commands to.
        """
        self.properties['repeat_commands'] = value

    @property
    def movement_speed(self):
        """
        Returns the current movement speed.
        :return: The current movement speed.
        :rtype: float
        """
        if 'movement_speed' in self.properties:
            return self.properties['movement_speed']
        else:
            return DEFAULT_MOVEMENT_SPEED

    @movement_speed.setter
    def movement_speed(self, value):
        """
        Sets the movement speed to the given value
        :param float value: The given value to set the movement speed to.
        """
        self.properties['movement_speed'] = value

    @property
    def default_control_stack(self):
        """
        :return: The default control stack for this object.
        :rtype: list
        """
        if 'default_control_stack' in self.properties:
            return eval(self.properties['default_control_stack'])
        else:
            return []

    @property
    def can_always_move(self):
        """
        :return: If true, this object can always move, even if its movement would normally be invalid.
        :rtype: bool
        """
        if 'can_always_move' in self.properties:
            return self.properties['can_always_move']
        else:
            return False

    @can_always_move.setter
    def can_always_move(self, value):
        """
        Changes the can_always_move value, allowing characters to move through anything.
        :param bool value: the new value
        """
        self.properties['can_always_move'] = value

    @property
    def default_interaction_stack(self):
        """
        :return: This object's default interaction stack (the list of control commands it should execute when interacted
        with unless otherwise specified by an event)
        :rtype: list
        """
        if "default_interaction_stack" in self.properties:
            return eval(self.properties["default_interaction_stack"])
        else:
            if "turn_on_interaction" in self.properties and self.properties['turn_on_interaction'] is False:
                return []
            else:
                return [ControlCommand("turn_to", target="pink_otm_current_pc", never_repeat=True)]


OrthogonalTiledMapGameObjectMobile.movement_commands = {
    "move_to": OrthogonalTiledMapGameObjectMobile.command_move_to,
    "move_left": OrthogonalTiledMapGameObjectMobile.command_move_left,
    "move_right": OrthogonalTiledMapGameObjectMobile.command_move_right,
    "move_up": OrthogonalTiledMapGameObjectMobile.command_move_up,
    "move_down": OrthogonalTiledMapGameObjectMobile.command_move_down,
    "move_random": OrthogonalTiledMapGameObjectMobile.command_move_random,
    "delay": OrthogonalTiledMapGameObjectMobile.command_delay,
    "change_movement_speed": OrthogonalTiledMapGameObjectMobile.command_change_movement_speed
}


class OrthogonalTiledMapGameObjectSpriteCollection(
    TiledMapGameObjectSpriteCollection, OrthogonalTiledMapGameObjectMobile
):
    movement_commands = {}

    def __init__(self, x, y, layer, properties, collection_path):
        """
        An object that uses a sprite collection on an Orthogonal Tiled Map.
        :param int x: Distance in pixels between the object's left side and the left border of the map.
        :param int y: Distance in pixels between the object's top side and the top border of the map.
        :param TiledMapGameLayer layer: The layer this object should be placed on.
        :param dict properties: Dictionary of pink engine properties that applies to this object.
        :param str collection_path: The path to the sprite collection to use.
        """
        if self.should_load_consistent(properties):
            old_incarnation = getattr(renpy.store, properties['ref_name'],
                                      None)  # type: OrthogonalTiledMapGameObjectSpriteCollection  # noqa
            self.sprite_collection_path = old_incarnation.sprite_collection_path
            self.sprite_collection = old_incarnation.sprite_collection
        else:
            self.sprite_collection_path = collection_path
            self.sprite_collection = TiledMapSpriteCollection(renpy.exports.file(collection_path))

        OrthogonalTiledMapGameObjectMobile.__init__(
            self, self.sprite_collection, x, y, self.sprite_collection.width, self.sprite_collection.height, layer,
            properties)
        TiledMapGameObjectSpriteCollection.__init__(self, x, y, layer, properties, collection_path)

        if type(self) is OrthogonalTiledMapGameObjectSpriteCollection and self.ref_name is not None:
            # Type comparison ensures this only happens at the end of the final loading
            setattr(renpy.store, self.ref_name, self)

    def _set_animation(self, animation_name, speed=None):
        """
        This function is called to determine the object's current animation set. It calls an oriented animation
        (ex. walk_left) if one is available.
        :param str animation_name: Name of new animation
        :param float speed: Speed at which the animation should be played.
        """
        TiledMapGameObjectSpriteCollection._set_animation(self, animation_name, speed)
        # if self.displayable_id is not None:  # updates display grid in case animation changes size
        #     self.parent.base_grid.move_element(self.displayable_id, self.x, self.y)
        #     TODO self.x and self.y are not valid because those are image coordinates, not base coordinates. Add
        #      recalc base function

    def _finish_executing_command(self):
        """
        Finishes executing the current ongoing control command.
        """
        OrthogonalTiledMapGameObjectMobile._finish_executing_command(self)

        if len(self.control_stack) == 0:
            self._set_animation(self.stand_animation)

    def set_to_x_y(self, x, y, override_prev_coord=True, **kwargs):
        """
        Instantly moves this object so that the left side of its image is x pixels from the left border of the map and
        the top side of its image is y pixels from the top right border of the map. Note that this overrides
        the saved x and saved y as well, so if used during a dynamic event, it will cause the object to not return
        to its starting location. Accepts the keyword argument 'orientation', which sets the new orienation.
        :param bool override_prev_coord: Whether to override the previous coordinate
        :param int x: The given x
        :param int y: The given y
        """
        if 'orientation' in kwargs and kwargs['orientation'] is not None:
            self._orientation = kwargs['orientation']
            self._set_animation(self.stand_animation)
        OrthogonalTiledMapGameObjectMobile.set_to_x_y(self, x, y, override_prev_coord, **kwargs)

    def should_be_auto_centered(self):
        """
        Whether this object should be auto-centered.
        :rtype: bool
        """
        return (
                renpy.store.pink_otm_autocenter_sprites and not self.forbid_autocenter and
                (self.base_width % self.parent.tile_size.x != 0 or self.base_height % self.parent.tile_size.y != 0) and
                self.x % self.parent.tile_size.x == 0 and (self.y + self.height) % self.parent.tile_size.y == 0
        )

    def in_place(self, x_coord=None, y_coord=None, orientation=None, finished_moving=True):
        """
        Returns True if the character is in the given location.
        :param int x_coord: the given x coordinate. None for wildcard.
        :param int y_coord: the given y coordinate. None for wildcard.
        :param str orientation: The given orientation. None for wildcard.
        :param bool finished_moving: Whether the control stack needs to be empty.
        :return:
        """
        if self.displayable_id is None:  # Not currently on the map for whatever reason
            return False
        if x_coord is not None and self.central_coord.x != x_coord:
            return False
        if y_coord is not None and self.central_coord.y != y_coord:
            return False
        if orientation is not None and self._orientation != orientation:
            return False
        if finished_moving and (self.control_stack != [] or self._current_command is not None):
            return False
        return True

    def switch_sprite_collection(self, target_sprite_collection):
        """
        Switches to a new sprite collection, ensuring that coordinates and animation name are kept the same.
        :param str target_sprite_collection: Path of the new sprite collection.
        """
        target_coord = self.central_coord

        self.parent.animated_tiles.remove(self.sprite_collection)
        animation_name = self.current_animation_name
        start_time = self.sprite_collection.latest_frame_gt
        self.parent.remove_element(self)
        self.current_animation_name = None  # So new animation doesn't register as the same
        self.sprite_collection = TiledMapSpriteCollection(renpy.exports.file(
            "pink_engine/sprite_collections/" + target_sprite_collection), start_time)
        self._set_animation(animation_name)
        self.parent.add_element(0, 0, self.layer, self)
        self.set_to_coords(target_coord.x, target_coord.y)

        self.parent.animated_tiles.append(self.sprite_collection)

    def freeze(self):
        """
        Freezes this object (causing it to remain motionless in place until unfrozen)
        """
        self.sprite_collection.pause_animation()
        OrthogonalTiledMapGameObjectMobile.freeze(self)

    def unfreeze(self, increment_gt=True):
        """
        Unfreezes this object
        :param bool increment_gt: Whether or not to increment the GT. When unfreezing on game load, this should not be
        done. Otherwise, it should be done.
        """
        self.sprite_collection.unpause_animation()
        if increment_gt:
            self.sprite_collection.increment_gt(self._last_gt)
        OrthogonalTiledMapGameObjectMobile.unfreeze(self, increment_gt)

    def force_stand(self):
        """
        Forces this object into its stand animation.
        """
        self._set_animation(self.stand_animation)

    def _command_move_direction(self, command, gt, target_x, target_y, pop=True):
        """
        Move a single square in a non-diagonal direction
        :param ControlCommand command: The control command which caused this movement
        :param float gt: The current game time in seconds
        :param int target_x: The x this object should have at the end of the given movement.
        :param int target_y: The y this object should have at the end of the given movement.
        :param bool pop: If True, removes the current top object from the control stack (affected by the command's
        pop_invalid_move value).
        """
        self._current_command = command

        # get movement_speed
        if command.movement_time is not None:
            movement_time = command.movement_time
        else:
            movement_time = self.movement_speed

        self._command_gt_start = gt
        self._command_gt_end = gt + movement_time
        self._command_x_start, self._command_y_start = self.x, self.y
        if self.can_always_move or self.parent.base_grid.is_move_permitted(self.displayable_id, target_x, target_y):
            self.prev_coord = self.central_coord
            self.parent.base_grid.move_element(self.displayable_id, target_x, target_y)
            self._set_animation(self.move_animation)
            self._command_x_end = target_x
            self._command_y_end = target_y
            if pop:
                self.finish_command(command)
        else:
            self._command_x_end, self._command_y_end = self.x, self.y
            if command.animate_invalid_move:
                self._set_animation(self.move_animation)
            else:
                self._set_animation(self.stand_animation)
            if pop and command.pop_invalid_move:
                self.finish_command(command)

    def command_delay(self, command, gt):
        """
        Executes a delay command, causing the object to wait a number of seconds until consuming its next command.
        :param ControlCommand command: The command being executed.
        :param float gt: The current game time in seconds
        """
        OrthogonalTiledMapGameObjectMobile.command_delay(self, command, gt)
        self._set_animation(self.stand_animation)

    def _command_turn_direction(self, command, gt, direction, pop=True):
        """
        Turn towards a given direction
        :param ControlCommand command: The command being executed.
        :param float gt: The current game time in seconds
        :param str direction: The new direction this character should face.
        :param bool pop: If True, removes the current top object from the control stack.
        """
        self._current_command = command

        self._orientation = direction
        if command.stand_on_turn:
            self._set_animation(self.stand_animation)

        self._current_command = command
        self._command_gt_start = gt
        self._command_gt_end = gt + self.movement_speed
        if pop:
            self.finish_command(command)

    def command_turn_left(self, command, gt, pop=True):
        """
        Executes a turn_left command
        :param ControlCommand command: The command being executed.
        :param float gt: The current game time in seconds
        :param bool pop: If True, removes the current top object from the control stack
        """
        self._command_turn_direction(command=command, gt=gt, direction="left", pop=pop)

    def command_turn_right(self, command, gt, pop=True):
        """
        Executes a turn_right command
        :param ControlCommand command: The command being executed.
        :param float gt: The current game time in seconds
        :param bool pop: If True, removes the current top object from the control stack
        """
        self._command_turn_direction(command=command, gt=gt, direction="right", pop=pop)

    def command_turn_up(self, command, gt, pop=True):
        """
        Executes a turn_up command
        :param ControlCommand command: The command being executed.
        :param float gt: The current game time in seconds
        :param bool pop: If True, removes the current top object from the control stack
        """
        self._command_turn_direction(command=command, gt=gt, direction="up", pop=pop)

    def command_turn_down(self, command, gt, pop=True):
        """
        Executes a turn_down command
        :param ControlCommand command: The command being executed.
        :param float gt: The current game time in seconds
        :param bool pop: If True, removes the current top object from the control stack
        """
        self._command_turn_direction(command=command, gt=gt, direction="down", pop=pop)

    def command_turn_clockwise(self, command, gt):
        """
        Executes and consumes a 'turn_clockwise' command, which causes the object to change its orientation in a
        clockwise manner.
        :param ControlCommand command: The command being executed.
        :param float gt: The current game time in seconds
        """
        if self._orientation == "up":
            self.command_turn_right(command, gt)
        elif self._orientation == "right":
            self.command_turn_down(command, gt)
        elif self._orientation == "down":
            self.command_turn_left(command, gt)
        elif self._orientation == "left":
            self.command_turn_up(command, gt)

    def command_turn_counter_clockwise(self, command, gt):
        """
        Executes and consumes a 'turn_counter_clockwise' command, which causes the object to change its orientation in a
        counter-clockwise manner.
        :param ControlCommand command: The command being executed.
        :param float gt: The current game time in seconds
        """
        if self._orientation == "up":
            self.command_turn_left(command, gt)
        elif self._orientation == "right":
            self.command_turn_up(command, gt)
        elif self._orientation == "down":
            self.command_turn_right(command, gt)
        elif self._orientation == "left":
            self.command_turn_down(command, gt)

    def command_turn_random(self, command, gt):
        """
        Executes and consumes a 'turn_random' command, which causes the object to change its orientation to a random
        direction.
        :param ControlCommand command: The command being executed.
        :param float gt: The current game time in seconds
        """
        directions = ["up", "down", "left", "right"]

        if command.exclude_current_direction and self._orientation in directions:
            directions.remove(self._orientation)

        random_direction = random.choice(directions)
        if random_direction == "up":
            self.command_turn_up(command, gt)
        elif random_direction == "right":
            self.command_turn_right(command, gt)
        elif random_direction == "down":
            self.command_turn_down(command, gt)
        elif random_direction == "left":
            self.command_turn_left(command, gt)

    def command_turn_to(self, command, gt):
        """
        Executes and consumes a 'turn_to' command, which causes the object to change its orientation to face a given
        coordinate or object.
        :param ControlCommand command: The command being executed.
        :param float gt: The current game time in seconds
        """
        target = command.target

        central_coord = self.central_coord
        if type(target) in [str, unicode]:  # noqa
            target_object = getattr(renpy.store, target)  # Py2
            if target_object is not None:
                target_coord = target_object.central_coord
                target = (target_coord.x, target_coord.y)
            else:
                self.finish_command(command)
                return

        x_diff = central_coord.x - target[0]
        y_diff = central_coord.y - target[1]
        if abs(x_diff) > abs(y_diff):
            if x_diff > 0:
                self.command_turn_left(command, gt)
            else:
                self.command_turn_right(command, gt)
        else:
            if y_diff > 0:
                self.command_turn_up(command, gt)
            else:
                self.command_turn_down(command, gt)

    def _command_go_direction(self, command, gt, direction, turn_func, move_func, pop=True):
        """
        Executes a go command for the given direction. Calls the appropriate turn command if not turned in the given
        direction, otherwise calls the movement command.
        :param ControlCommand command: The command being executed.
        :param float gt: The current game time in seconds
        :param str direction: The given direction
        :param func turn_func: The relevant turn function
        :param func move_func: The relevant movement function
        :param bool pop:
        """
        if command.turn_and_move:
            if self._orientation != direction:
                turn_func(command, gt, pop=False)
            move_func(command, gt, pop=pop)
        else:
            if self._orientation != direction:
                turn_func(command, gt, pop=pop)
            else:
                move_func(command, gt, pop=pop)

    def command_go_left(self, command, gt, pop=True):
        """
        Executes a go_left command. Calls turn_left if the character is not oriented leftwards, and move_left
        otherwise. Does not consume the command, as that happens in its called functions.

        :param ControlCommand command: The command being executed.
        :param float gt: The current game time in seconds
        :param bool pop: Whether or not to pop the top command on the command stack. Disabled if the turning is
        part of a broader command.
        """
        self._command_go_direction(
            command=command, gt=gt, direction="left", turn_func=self.command_turn_left,
            move_func=self.command_move_left, pop=pop)

    def command_go_right(self, command, gt, pop=True):
        """
        Executes a go_right command. Calls turn_right if the character is not oriented rightwards, and move_right
        otherwise. Does not consume the command, as that happens in its called functions.

        :param ControlCommand command: The command being executed.
        :param float gt: The game timebase, in seconds.
        :param bool pop: Whether or not to pop the top command on the command stack. Disabled if the turning is
        part of a broader command.
        """
        self._command_go_direction(
            command=command, gt=gt, direction="right", turn_func=self.command_turn_right,
            move_func=self.command_move_right, pop=pop)

    def command_go_up(self, command, gt, pop=True):
        """
        Executes a go_up command. Calls turn_up if the character is not oriented upwards, and move_up
        otherwise. Does not consume the command, as that happens in its called functions.

        :param ControlCommand command: The command being executed.
        :param float gt: The current game time in seconds
        :param bool pop: Whether or not to pop the top command on the command stack. Disabled if the turning is
        part of a broader command.
        """
        self._command_go_direction(
            command=command, gt=gt, direction="up", turn_func=self.command_turn_up,
            move_func=self.command_move_up, pop=pop)

    def command_go_down(self, command, gt, pop=True):
        """
        Executes a go_down command. Calls turn_down if the character is not oriented downwards, and move_down
        otherwise. Does not consume the command, as that happens in its called functions.

        :param ControlCommand command: The command being executed.
        :param float gt: The current game time in seconds
        :param bool pop: Whether or not to pop the top command on the command stack. Disabled if the turning is
        part of a broader command.
        """
        self._command_go_direction(
            command=command, gt=gt, direction="down", turn_func=self.command_turn_down,
            move_func=self.command_move_down, pop=pop)

    def command_go_to(self, command, gt):
        """
        Executes and consumes a go_to command, causing an object to move towards a location. The command isn't
        finished until the object has reached the location

        :param ControlCommand command: The command being executed
        :param float gt: The current game time in seconds
        """
        target = command.target

        central_coord = self.central_coord
        if type(target) in [str, unicode]:  # noqa
            target = 'renpy.store.' + target
            target_object = eval(target)
            if target_object is not None:
                if command.prev_coord and target_object.prev_coord is not None:
                    target_coord = target_object.prev_coord
                else:
                    target_coord = target_object.central_coord
                target = (target_coord.x, target_coord.y)
            else:
                self.finish_command(command)

        if central_coord.y < target[1]:
            self.command_go_down(command, gt, pop=False)
        elif central_coord.y > target[1]:
            self.command_go_up(command, gt, pop=False)
        elif central_coord.x < target[0]:
            self.command_go_right(command, gt, pop=False)
        elif central_coord.x > target[0]:
            self.command_go_left(command, gt, pop=False)
        else:
            self._set_animation(self.stand_animation)

        # Updates coords after movement
        central_coord = self.central_coord

        # If single step is disabled, then the command isn't consumed until the object reaches the target.
        if (central_coord.x, central_coord.y) == target or command.single_step:
            self.finish_command(command)

    def command_go_random(self, command, gt):
        """
        Executes a go_random command.

        :param ControlCommand command: The command being executed.
        :param float gt: The current game time in seconds
        """
        directions = ["up", "down", "left", "right"]

        for coord in self.coords:
            if "right" in directions and coord.x >= command.max_x:
                directions.remove("right")
            if "left" in directions and coord.x <= command.min_x:
                directions.remove("left")
            if "down" in directions and coord.y >= command.max_y:
                directions.remove("down")
            if "up" in directions and coord.y <= command.min_y:
                directions.remove("up")

        if len(directions) == 0:
            # If no movement is permitted.
            self.finish_command(command)
        else:
            random_direction = random.choice(directions)

            if random_direction == "up":
                self.command_go_up(command, gt)
            elif random_direction == "down":
                self.command_go_down(command, gt)
            elif random_direction == "left":
                self.command_go_left(command, gt)
            if random_direction == "right":
                self.command_go_right(command, gt)

    def command_go_custom(self, command, gt):  # noqa
        """
        Executes and consumes a go_custom command. Causes the character to move a certain amount of pixels in both the
        x and y directions. Used for gridless movement.
        :param ControlCommand command: The command being executed.
        :param float gt: The current game time in seconds
        """
        move_by_x = command.move_by_x
        move_by_y = command.move_by_y
        if move_by_x != 0:
            move_x_to_y = float(move_by_y) / move_by_x
        else:
            move_x_to_y = 0.0

        # Variables that only have to be calculated once
        base_x_start = self.base_x_start
        base_x_end = self.base_x_end
        base_y_start = self.base_y_start
        base_y_end = self.base_y_end
        movement_speed = self.movement_speed

        # Coordinates at which object crosses a grid line.
        cross_coordinates = []

        if move_by_x > 0:
            # Right side of the base crosses grid lines first.
            start_x = self.parent.pixel_to_coord(base_x_end, is_x=True)
            end_x = self.parent.pixel_to_coord(base_x_end + move_by_x, is_x=True)

            previous_coord = Coord(base_x_start, base_y_start)
            base_width = self.base_width
            for x_coord in range(start_x, end_x):
                # crossing at x border x_coord x_coord + 1
                cross_x = ((x_coord + 1) * self.parent.tile_size.x) - base_width
                cross_y = previous_coord.y + int((cross_x - previous_coord.x) * move_x_to_y)
                cross_coord = Coord(cross_x, cross_y)
                previous_coord = cross_coord
                cross_coordinates.append(cross_coord)
        elif move_by_x < 0:
            # Left side of the base crosses grid lines first.
            start_x = self.parent.pixel_to_coord(base_x_start, is_x=True)
            end_x = self.parent.pixel_to_coord(base_x_start + move_by_x, is_x=True)

            previous_coord = Coord(base_x_start, base_y_start)
            for x_coord in range(start_x, end_x, -1):
                # crossing at x border x_coord x_coord - 1
                cross_x = x_coord * self.parent.tile_size.x
                cross_y = previous_coord.y + int((cross_x - previous_coord.x) * move_x_to_y)
                cross_coord = Coord(cross_x, cross_y)
                previous_coord = cross_coord
                cross_coordinates.append(cross_coord)

        if move_by_y > 0:
            # Bottom side of the base crossed grid lines first.
            start_y = self.parent.pixel_to_coord(base_y_end, is_x=False)
            end_y = self.parent.pixel_to_coord(base_y_end + move_by_y, is_x=False)

            previous_coord = Coord(base_x_start, base_y_start)
            base_height = self.base_height
            for y_coord in range(start_y, end_y):
                # crossing at y border y_coord y_coord - 1
                cross_y = ((y_coord + 1) * self.parent.tile_size.y) - base_height
                cross_x = previous_coord.x + int((cross_y - previous_coord.y) / move_x_to_y)
                cross_coord = Coord(cross_x, cross_y)
                previous_coord = cross_coord
                if cross_coord not in cross_coordinates:
                    cross_coordinates.append(cross_coord)
        elif move_by_y < 0:
            # Top size of the base crosses grid lines first.
            start_y = self.parent.pixel_to_coord(base_y_start, is_x=False)
            end_y = self.parent.pixel_to_coord(base_y_start + move_by_y, is_x=False)

            previous_coord = Coord(base_x_start, base_y_start)
            for y_coord in range(start_y, end_y, -1):
                # crossing at y border y_coord y_coord + 1
                cross_y = (y_coord * self.parent.tile_size.y)
                cross_x = previous_coord.x + int((cross_y - previous_coord.y) / move_x_to_y)
                cross_coord = Coord(cross_x, cross_y)
                previous_coord = cross_coord
                if cross_coord not in cross_coordinates:
                    cross_coordinates.append(cross_coord)

        # Sort crossing coordinates by distance
        if move_by_x > 0:
            cross_coordinates = sorted(cross_coordinates, key=lambda z: z[0])
        elif move_by_x < 0:
            cross_coordinates = sorted(cross_coordinates, key=lambda z: z[0], reverse=True)
        elif move_by_y > 0:
            cross_coordinates = sorted(cross_coordinates, key=lambda z: z[1])
        elif move_by_y < 0:
            cross_coordinates = sorted(cross_coordinates, key=lambda z: z[1], reverse=True)

        # Adds final, non-grid crossing, end coordinate
        final_coordinate = Coord(base_x_start + move_by_x, base_y_start + move_by_y)
        if final_coordinate not in cross_coordinates:
            cross_coordinates.append(final_coordinate)

        previous_coord = Coord(base_x_start, base_y_start)
        command_components = []
        for coordinate in cross_coordinates:
            comp_move_by_x = coordinate.x - previous_coord.x
            comp_move_by_y = coordinate.y - previous_coord.y
            if comp_move_by_x == 0 and comp_move_by_y == 0:
                continue

            # Seconds to cross a pixel horizontally
            horizontal_speed = movement_speed / self.parent.tile_size.x
            # Seconds to cross a pixel vertically
            vertical_speed = movement_speed / self.parent.tile_size.y

            horizontal_time = horizontal_speed * comp_move_by_x
            vertical_time = vertical_speed * comp_move_by_y

            # Seconds necessary for the movement
            diagonal_time = sqrt((horizontal_time * horizontal_time) + (vertical_time * vertical_time))

            command_component = ControlCommand(
                "custom_component", move_by_x=comp_move_by_x, move_by_y=comp_move_by_y,
                pop_invalid_move=command.pop_invalid_move, animate_invalid_move=command.animate_invalid_move,
                never_repeat=True, movement_time=diagonal_time)
            command_components.append(command_component)
            previous_coord = coordinate

        # Removes this command
        self.finish_command(command)

        # Adds component commands.
        for x in range(len(command_components)):
            self.control_stack.insert(x, command_components[x])

    def command_custom_component(self, command, gt):
        """
        Executes and consumes a custom_component command. Custom movements (non-grid movements) are split into these,
        with each custom component consisting of at most a single grid coordinate change.

        :param ControlCommand command: The command being executed.
        :param float gt: The current game time in seconds
        """
        # sets orientation
        self._current_command = command
        move_by_x = command.move_by_x
        move_by_y = command.move_by_y
        if abs(move_by_x) > abs(move_by_y):
            if move_by_x > 0:
                self._orientation = "right"
            else:
                self._orientation = "left"
        else:
            if move_by_y > 0:
                self._orientation = "down"
            elif move_by_y < 0:
                # This is elif rather than else to prevent turning if both move_by_x and move_by_y end up being 0 for
                # some reason.
                self._orientation = "up"

        # Check if permitted
        target_x = self.x + move_by_x
        target_y = self.y + move_by_y
        permitted_move = True
        if not (
                self.can_always_move or
                self.parent.base_grid.is_move_permitted(self.displayable_id, target_x, target_y)):
            permitted_move = False

        # get movement_speed
        if command.movement_time is not None:
            movement_time = command.movement_time
        else:
            movement_time = self.movement_speed

        # Sets movement parameters
        if permitted_move:
            self.parent.base_grid.move_element(self.displayable_id, target_x, target_y)
            self._command_gt_start = gt
            self._command_gt_end = gt + movement_time
            self._command_x_start, self._command_y_start = self.x, self.y
            self._command_x_end = target_x
            self._command_y_end = target_y

            self._set_animation(self.move_animation)
            self.finish_command(command)
        else:
            self._command_gt_start = gt
            self._command_gt_end = gt + self.movement_speed
            if command.animate_invalid_move:
                self._set_animation(self.move_animation)
            else:
                self._set_animation(self.stand_animation)
            if command.pop_invalid_move:
                self.finish_command(command)

    def command_go_to_custom(self, command, gt):  # noqa
        """
        Executes and consumes a go_custom command. Causes the character to move the top-left coordinate of their
        base to a given pixel.
        :param ControlCommand command: The command being executed.
        :param float gt: The current game time in seconds
        """
        target = command.target
        move_by_x = target[0] - self.base_x_start
        move_by_y = target[1] - self.base_y_end

        go_to_command = ControlCommand(
            "go_custom", move_by_x=move_by_x, move_by_y=move_by_y,
            pop_invalid_move=command.pop_invalid_move, animate_invalid_move=command.animate_invalid_move,
            never_repeat=True)
        self.finish_command(command)
        self.control_stack.insert(0, go_to_command)

    def command_play_animation(self, command, gt):
        """
        Executes and consumes a 'play_animation' command, which causes a single animation sequence for the current
        sprite collection to play in full. The next command isn't executed until the animation is finished playing.

        :param ControlCommand command: The command being executed.
        :param float gt: The current game time in seconds
        """
        self._current_command = command

        if command.animation_speed is not None:
            self._set_animation(command.animation_name, speed=command.animation_speed)
        else:
            self._set_animation(command.animation_name)

        animation_duration = (
                (len(self.sprite_collection.current_animation.images) * self.sprite_collection.current_speed) - 0.05
        )
        # The 0.05 is for situations wherein multiple animations/animation changes are placed after each other,
        # preventing the tiny frame of stand/move animation that would otherwise occur between these.

        self._command_gt_start = gt
        self._command_gt_end = gt + animation_duration

        self.finish_command(command)

    def command_change_stand_animation(self, command, gt):
        """
        Executes and consumes a 'change_stand_animation' command, which changes the stand animation for this object.

        :param ControlCommand command: The command being executed.
        :param float gt: The current game time in seconds
        """
        self._current_command = command

        currently_standing = False
        if self.current_animation_name == self.stand_animation:
            currently_standing = True

        self.stand_animation = command.animation_name

        if currently_standing:
            self._set_animation(self.stand_animation)

        self._command_gt_start = gt
        self._command_gt_end = gt

        self.finish_command(command)

    def command_change_move_animation(self, command, gt):
        """
        Executes and consumes a 'change_move_animation' command, which changes the move animation for this object.

        :param ControlCommand command: The command being executed.
        :param float gt: The current game time in seconds
        """
        self._current_command = command

        currently_moving = False
        if self.current_animation_name == self.move_animation:
            currently_moving = True
        self.move_animation = command.animation_name
        if currently_moving:
            self._set_animation(self.move_animation)

        if self is self.parent.player_object:
            for follower in renpy.store.pink_otm_followers:
                if follower.current_animation_name == follower.move_animation:
                    currently_moving = True
                follower.move_animation = command.animation_name
                if currently_moving:
                    follower._set_animation(follower.move_animation)   # noqa

        self._command_gt_start = gt
        self._command_gt_end = gt

        self.finish_command(command)

    @property
    def height(self):
        """
        :return: This object's image height in pixel
        :rtype: int
        """
        return self.sprite_collection.height

    @property
    def width(self):
        """
        :return: This object's image width in pixel
        :rtype: int
        """
        return self.sprite_collection.width

    @property
    def base_offset_x_start(self):
        """
        :return: The x offset of the start of the object's base (where the object begins for mapping and event
        purposes in relation to where the image of the object begins)
        :rtype: int
        """
        if "base_offset_x_start" in self.sprite_collection.current_animation.properties:
            return int(self.sprite_collection.current_animation.properties["base_offset_x_start"] * self.xzoom)
        else:
            return 0

    @property
    def base_offset_x_end(self):
        """
        :return: The x offset of the end of the object's base (where the object ends for mapping and event
        purposes in relation to where the image of the object begins)
        :rtype: int
        """
        if "base_offset_x_end" in self.sprite_collection.current_animation.properties:
            return int(self.sprite_collection.current_animation.properties["base_offset_x_end"] * self.xzoom)
        else:
            return self.sprite_collection.current_animation.image_width

    @property
    def base_offset_y_start(self):
        """
        :return: The y offset of the start of the object's base (where the object begins for mapping and event
        purposes in relation to where the image of the object begins)
        :rtype: int
        """
        if "base_offset_y_start" in self.sprite_collection.current_animation.properties:
            return int(self.sprite_collection.current_animation.properties["base_offset_y_start"] * self.yzoom)
        else:
            return 0

    @property
    def base_offset_y_end(self):
        """
        :return: The y offset of the end of the object's base (where the object ends for mapping and event
        purposes in relation to where the image of the object begins)
        :rtype: int
        """
        if "base_offset_y_end" in self.sprite_collection.current_animation.properties:
            return int(self.sprite_collection.current_animation.properties["base_offset_y_end"] * self.yzoom)
        else:
            return self.sprite_collection.current_animation.image_height

    @property
    def move_from(self):
        """
        :return: The movement allowed from a square occupied by this object.
        :rtype: str|None
        """
        if "move_from" in self.properties:
            return self.properties["move_from"]
        elif "move_from" in self.sprite_collection.current_animation.properties:
            return self.sprite_collection.current_animation.properties["move_from"]
        else:
            return None

    @property
    def move_to(self):
        """
        :return: The movement allowed to a square occupied by this object.
        :rtype: str|None
        """
        if "move_to" in self.properties:
            return self.properties["move_to"]
        elif "move_to" in self.sprite_collection.current_animation.properties:
            return self.sprite_collection.current_animation.properties["move_to"]
        else:
            return None

    @property
    def forbid_autocenter(self):
        """
        :return: Whether autocenter settings should be overwritten for this sprite collection.
        :rtype: bool
        """
        if 'forbid_autocenter' in self.properties:
            return self.properties['forbid_autocenter']
        else:
            return False


OrthogonalTiledMapGameObjectSpriteCollection.movement_commands = OrthogonalTiledMapGameObjectMobile.movement_commands.copy()  # noqa
OrthogonalTiledMapGameObjectSpriteCollection.movement_commands.update({
    "delay": OrthogonalTiledMapGameObjectSpriteCollection.command_delay,
    "go_to": OrthogonalTiledMapGameObjectSpriteCollection.command_go_to,
    "go_left": OrthogonalTiledMapGameObjectSpriteCollection.command_go_left,
    "go_right": OrthogonalTiledMapGameObjectSpriteCollection.command_go_right,
    "go_up": OrthogonalTiledMapGameObjectSpriteCollection.command_go_up,
    "go_down": OrthogonalTiledMapGameObjectSpriteCollection.command_go_down,
    "go_random": OrthogonalTiledMapGameObjectSpriteCollection.command_go_random,
    "go_custom": OrthogonalTiledMapGameObjectSpriteCollection.command_go_custom,
    "custom_component": OrthogonalTiledMapGameObjectSpriteCollection.command_custom_component,
    "go_to_custom": OrthogonalTiledMapGameObjectSpriteCollection.command_go_to_custom,
    "turn_left": OrthogonalTiledMapGameObjectSpriteCollection.command_turn_left,
    "turn_right": OrthogonalTiledMapGameObjectSpriteCollection.command_turn_right,
    "turn_up": OrthogonalTiledMapGameObjectSpriteCollection.command_turn_up,
    "turn_down": OrthogonalTiledMapGameObjectSpriteCollection.command_turn_down,
    "turn_clockwise": OrthogonalTiledMapGameObjectSpriteCollection.command_turn_clockwise,
    "turn_counter_clockwise": OrthogonalTiledMapGameObjectSpriteCollection.command_turn_counter_clockwise,
    "turn_random": OrthogonalTiledMapGameObjectSpriteCollection.command_turn_random,
    "turn_to": OrthogonalTiledMapGameObjectSpriteCollection.command_turn_to,
    "play_animation": OrthogonalTiledMapGameObjectSpriteCollection.command_play_animation,
    "change_stand_animation": OrthogonalTiledMapGameObjectSpriteCollection.command_change_stand_animation,
    "change_move_animation": OrthogonalTiledMapGameObjectSpriteCollection.command_change_move_animation
})


class OrthogonalTiledMapTiledBasesEntry(object):
    def __init__(self, parent_map, element_id, x, y, layer, game_object):
        """
        A single object's base.
        :param OrthogonalTiledMap parent_map: The game map this base is a part of.
        :param int element_id: The element id of this base. Should be the same as the displayable id for the object
        this is the base of, if this base is linked to an object. If not, this id should be negative.
        :param int x: Distance in pixels between this base's object's left side and the left border of the map.
        :param int y: Distance in pixels between this base's object's top side and the top border of the map.
        :param TiledMapGameLayer layer: The later this object is placed on
        :param OrthogonalTiledMapGameObjectBase game_object: The game object this is the base of
        """
        self.parent = parent_map
        self.game_object = game_object
        self.element_id = element_id

        self.base_x_start = x + game_object.base_offset_x_start
        self.base_y_start = y + game_object.base_offset_y_start
        self.base_x_end = x + game_object.base_offset_x_end
        self.base_y_end = y + game_object.base_offset_y_end

        self.base_priority = (LAYER_PRIORITY[layer.layer_type], layer.z_order, self.element_id)

    def recalculate_base(self, new_x, new_y):
        """
        Recalculate's this base for a new x and y
        :param int new_x: The new distance in pixels between this base's object's left side and the left border of the
        map.
        :param int new_y: The new distance in pixels between this base's object's top side and the top border of the
        map.
        """
        self.base_x_start = new_x + self.game_object.base_offset_x_start
        self.base_y_start = new_y + self.game_object.base_offset_y_start
        self.base_x_end = new_x + self.game_object.base_offset_x_end
        self.base_y_end = new_y + self.game_object.base_offset_y_end

    def get_base_columns(self):
        """
        :return: The range of indexes of columns this base intersects.
        :rtype: range
        """
        start_x_coord = int(self.base_x_start / self.parent.tile_size.x)  # BASE NO WORK FOR TILE.
        end_x_coord = int((self.base_x_end - 1) / self.parent.tile_size.x)  # BASE NO WORK FOR TILE.
        return range(start_x_coord, end_x_coord + 1)  # BASE NO WORK FOR TILE.

    def get_base_rows(self):
        """
        :return: The range of indexes of rows this base intersects.
        :rtype: range
        """
        start_y_coord = int(self.base_y_start / self.parent.tile_size.y)
        end_y_coord = int((self.base_y_end - 1) / self.parent.tile_size.y)
        return range(start_y_coord, end_y_coord + 1)

    @property
    def central_coord(self):
        """
        :return: The coordinate covered by the center of this object's base.
        :rtype: Coord
        """
        tile_width = self.parent.tile_size.x
        tile_height = self.parent.tile_size.y

        x_coord = int(((self.base_x_start + self.base_x_end - 1) / 2) / tile_width)
        y_coord = int(((self.base_y_start + self.base_y_end - 1) / 2) / tile_height)

        return Coord(x_coord, y_coord)

    @property
    def coords(self):
        """
        :return: The list of coordinates covered by this object's base.
        :rtype: list
        """
        start_x_coord = int(self.base_x_start / self.parent.tile_size.x)
        end_x_coord = int((self.base_x_end - 1) / self.parent.tile_size.x)
        start_y_coord = int(self.base_y_start / self.parent.tile_size.y)
        end_y_coord = int((self.base_y_end - 1) / self.parent.tile_size.y)

        coords = []
        for x in range(start_x_coord, end_x_coord + 1):
            for y in range(start_y_coord, end_y_coord + 1):
                coords.append(Coord(x, y))

        return coords


class OrthogonalTiledMapTiledBases(object):
    def __init__(self, game_object, x_dim, y_dim):
        """
        The structured collection of bases for an orthogonal tiled map
        :param OrthogonalTiledMap game_object: The game map this is the base collection of.
        :param int x_dim: The width in pixels of the map.
        :param int y_dim: The height in pixels of the map.
        """
        self.game_object = game_object

        # Creates a grid for all objects, to more efficiently check for objects that might be in your way.
        self.base_grid = []
        self.base_grid_x_size = x_dim
        self.base_grid_y_size = y_dim
        for x in range(self.base_grid_x_size):
            self.base_grid.append([])
            for y in range(self.base_grid_y_size):
                self.base_grid[x].append({})

        # Index of all bases
        self._next_id = -1  # type: int  # Negative numbers are used by tiles. Objects use displayable ids.
        self.map_elements = {}

    def get_next_id(self):
        """
        :return: The next ID that should be given to a tile.
        :rtype: int
        """
        next_id = self._next_id
        self._next_id -= 1
        return next_id

    def add_element(self, x, y, layer, game_object):
        """
        Adds the base for the given game_object to
        :param int x: Distance in pixels between the object's left side and the left border of the map.
        :param int y: Distance in pixels between the object's top side and the top border of the map.
        :param TiledMapGameLayer layer: The layer the object is placed on
        :param OrthogonalTiledMapGameTile|OrthogonalTiledMapGameObject game_object:
        """
        if game_object.single_display and game_object.displayable_id is not None:
            element_id = game_object.displayable_id
        else:
            element_id = self.get_next_id()
        new_element = OrthogonalTiledMapTiledBasesEntry(self.game_object, element_id, x, y, layer, game_object)
        self.map_elements[element_id] = new_element

        self.add_to_grid(new_element)

    def remove_element(self, element_id):
        """
        Removes the element with the given id.
        :param int element_id: The given id
        """
        element = self.map_elements.get(element_id)

        if element is not None:
            self.remove_from_grid(element)

    def add_to_grid(self, element):
        """
        Adds the given base to this collection
        :param OrthogonalBaseEntry element: The given base
        """
        for column in element.get_base_columns():
            for row in element.get_base_rows():
                if 0 <= row < self.base_grid_y_size and 0 <= column < self.base_grid_x_size:
                    self.base_grid[column][row][element.base_priority] = element

    def remove_from_grid(self, element):
        """
        Removes the given base from this collection
        :param OrthogonalBaseEntry element: the given base
        """
        for column in element.get_base_columns():
            for row in element.get_base_rows():
                if 0 <= row < self.base_grid_y_size and 0 <= column < self.base_grid_x_size:
                    self.base_grid[column][row].pop(element.base_priority)

    def move_element(self, element_id, new_x, new_y):
        """
        Moves the element with the given id to the given x and y
        :param int element_id: the given element id
        :param int new_x: the given x
        :param int new_y: the given y
        """
        element = self.map_elements[element_id]
        self.remove_from_grid(element)
        element.recalculate_base(new_x, new_y)
        self.add_to_grid(element)

    def get_top_base_with_property(
            self, base_property, x_coord, y_coord, ignore_element=None, default_value=None, additional_requirements=None
    ):
        """
        For the given property name, retrieves the value of the highest base where the value is set, and which
        intersects with the given coordinate.
        :param str base_property: The given property name
        :param int x_coord: The given x coordinate
        :param int y_coord: The given y coordinate
        :param OrthogonalBaseEntry ignore_element: Ignore this element in the retrieval process (used to prevent objects
        from preventing their own movement and such)
        :param list additional_requirements:
        :param default_value: The value that should be returned if no object whose base intersects with the given
        coordinates has the given property set.
        """
        objects_at_coord = self.base_grid[x_coord][y_coord].copy()

        # Removes ignored element
        if ignore_element is not None:
            ignore_base_priority = ignore_element.base_priority
            if ignore_base_priority in objects_at_coord:
                objects_at_coord.pop(ignore_base_priority)

        # Gets property
        matching_property = None
        corresponding_object = None
        for base_object_priority in sorted(objects_at_coord.keys(), reverse=True):
            base_object = objects_at_coord[base_object_priority]
            if additional_requirements is not None:
                additional_requirements_met = True

                for additional_requirement in additional_requirements:
                    if not getattr(base_object.game_object, additional_requirement):
                        additional_requirements_met = False
                        break

                if not additional_requirements_met:
                    break

            base_object_property = getattr(base_object.game_object, base_property)
            if base_object_property is not None:
                corresponding_object = base_object.game_object
                matching_property = base_object_property
                break

        # If no rules are found:
        if matching_property is None:
            matching_property = default_value

        return matching_property, corresponding_object

    def get_activation_event(self, x_coord, y_coord, ignore_element=None):
        return self.get_top_base_with_property('event_on_activate', x_coord, y_coord, ignore_element=ignore_element,
                                               additional_requirements=['event_conditional_met'])

    def get_touch_event(self, x_coord, y_coord, ignore_element=None):
        return self.get_top_base_with_property('event_on_touch', x_coord, y_coord, ignore_element=ignore_element,
                                               additional_requirements=['event_conditional_met'])

    def get_activation_code(self, x_coord, y_coord, ignore_element=None):
        return self.get_top_base_with_property('code_on_activate', x_coord, y_coord, ignore_element=ignore_element,
                                               additional_requirements=['code_conditional_met'])

    def get_touch_code(self, x_coord, y_coord, ignore_element=None):
        return self.get_top_base_with_property('code_on_touch', x_coord, y_coord, ignore_element=ignore_element,
                                               additional_requirements=['code_conditional_met'])

    def base_to_coords(self, base_x_start, base_x_end, base_y_start, base_y_end):
        """
        :param int base_x_start: The distance in pixels between the left side of the base and the left border of the
        map.
        :param int base_x_end: The distance in pixels between the right side of the base and the left border of the
        map.
        :param int base_y_start: The distance in pixels between the top side of the base and the upper border of the
        map.
        :param int base_y_end: The distance in pixels between the bottom side of the base and the upper border of the
        map.
        :return: The list of coordinates that a base with the given start and end pixels would occupy.
        :rtype: list
        """
        start_x_coord = int(base_x_start / self.game_object.tile_size.x)
        end_x_coord = int((base_x_end - 1) / self.game_object.tile_size.x)
        start_y_coord = int(base_y_start / self.game_object.tile_size.y)
        end_y_coord = int((base_y_end - 1) / self.game_object.tile_size.y)

        coords = []
        for x in range(start_x_coord, end_x_coord + 1):
            for y in range(start_y_coord, end_y_coord + 1):
                coords.append(Coord(x, y))

        return coords

    def _check_movement_straight(self, old_coords, new_coords, ignore_element, relevant_from_side, relevant_to_side):
        """
        :param list old_coords: list of old coordinate tuples
        :param list new_coords: list of new coordinate tuples
        :param OrthogonalBaseEntry ignore_element: Base to ignore (used to make objects not stand in the way of their
        own movement)
        :param int relevant_from_side: The index of the movement rule to look at for the movement from the
         old tile. (0 is top, 1 is right, 2 is bottom, 3 is left)
        :param int relevant_to_side: The index of the movement rule to look at for the movement to the
         new tile. (0 is top, 1 is right, 2 is bottom, 3 is left)
        :return: Whether or not straight movement between the given set of old coordinates and the given set of new
        coordinates would be allowed.
        :rtype: bool
        """
        for x, y in old_coords:
            from_rules, rules_object = self.get_top_base_with_property(
                'move_from', x, y, ignore_element=ignore_element, default_value='1111')
            if from_rules[relevant_from_side] == "0":
                return False
        for x, y in new_coords:
            to_rules, rules_object = self.get_top_base_with_property(
                'move_to', x, y, ignore_element=ignore_element, default_value='0000')
            if to_rules[relevant_to_side] == "0":
                return False
        return True

    def _check_movement_diagonal(self, old_coords, new_coords, ignore_element, relevant_from_sides, relevant_to_sides):
        """
        :param list old_coords: list of old coordinate tuples
        :param list new_coords: list of new coordinate tuples
        :param OrthogonalBaseEntry ignore_element: Base to ignore (used to make objects not stand in the way of their
        own movement)
        :param list relevant_from_sides: The indices of the movement rule to look at for the movement from the
         old tile. (0 is top, 1 is right, 2 is bottom, 3 is left)
        :param list relevant_to_sides: The indices of the movement rule to look at for the movement to the
         new tile. (0 is top, 1 is right, 2 is bottom, 3 is left)
        :return: Whether or not diagonal movement between the given set of old coordinates and the given set of new
        coordinates would be allowed.
        :rtype: bool
        """
        for x, y in old_coords:
            from_rules, rules_object = self.get_top_base_with_property(
                'move_from', x, y, ignore_element=ignore_element, default_value='1111')
            for relevant_from_side in relevant_from_sides:
                if from_rules[relevant_from_side] == "0":
                    return False
        for x, y in new_coords:
            to_rules, rules_object = self.get_top_base_with_property(
                'move_to', x, y, ignore_element=ignore_element, default_value='0000')
            for relevant_to_side in relevant_to_sides:
                if to_rules[relevant_to_side] == "0":
                    return False
        return True

    def is_move_permitted(self, element_id, new_x, new_y):
        """
        :param int element_id: The id of the object for which to check.
        :param int new_x: the given new x
        :param int new_y: the given new y
        :return: whether or not the base with the given ID is allowed to move to the given x and y.
        :rtype: bool
        """
        element = self.map_elements[element_id]
        x_diff = new_x - element.game_object.x
        y_diff = new_y - element.game_object.y

        old_coords = element.coords
        new_coords = self.base_to_coords(
            element.base_x_start + x_diff, element.base_x_end + x_diff, element.base_y_start + y_diff,
            element.base_y_end + y_diff)

        # Moving outside the map is always illegal
        for new_x, new_y in new_coords:
            if new_x < 0 or new_x > self.game_object.grid_size.x - 1:
                return False
            if new_y < 0 or new_y > self.game_object.grid_size.y - 1:
                return False

        x_coord_diff = new_coords[0].x - old_coords[0].x
        y_coord_diff = new_coords[0].y - old_coords[0].y

        # moving within your current coordinate is always permitted
        if x_coord_diff == 0 and y_coord_diff == 0:
            return True

        # Checking movement for coordinates
        if x_coord_diff > 0 and y_coord_diff == 0:  # right
            return self._check_movement_straight(
                old_coords, new_coords, ignore_element=element, relevant_from_side=1, relevant_to_side=3)
        elif x_coord_diff < 0 and y_coord_diff == 0:  # left
            return self._check_movement_straight(
                old_coords, new_coords, ignore_element=element, relevant_from_side=3, relevant_to_side=1)
        elif x_coord_diff == 0 and y_coord_diff < 0:  # up
            return self._check_movement_straight(
                old_coords, new_coords, ignore_element=element, relevant_from_side=0, relevant_to_side=2)
        elif x_coord_diff == 0 and y_coord_diff > 0:  # down
            return self._check_movement_straight(
                old_coords, new_coords, ignore_element=element, relevant_from_side=2, relevant_to_side=0)
        elif x_coord_diff > 0 and y_coord_diff > 0:  # right-down
            return self._check_movement_diagonal(
                old_coords, new_coords, ignore_element=element, relevant_from_sides=[1, 2], relevant_to_sides=[3, 0])
        elif x_coord_diff > 0 and y_coord_diff < 0:  # noqa  # right-up
            return self._check_movement_diagonal(
                old_coords, new_coords, ignore_element=element, relevant_from_sides=[1, 0], relevant_to_sides=[3, 2])
        elif x_coord_diff < 0 and y_coord_diff > 0:  # noqa  # left-down
            return self._check_movement_diagonal(
                old_coords, new_coords, ignore_element=element, relevant_from_sides=[3, 2], relevant_to_sides=[1, 0])
        elif x_coord_diff < 0 and y_coord_diff < 0:  # noqa  # left-up
            return self._check_movement_diagonal(
                old_coords, new_coords, ignore_element=element, relevant_from_sides=[3, 0], relevant_to_sides=[1, 2])

        return False


class OrthogonalTiledCamera(TiledMapCamera):
    def __init__(self, game_map):
        """
        This class serves as the in-game camera for the Pink Engine maps, with the x and y being used as the
        viewport adjustment variables in Pink Engine map screens.
        :param OrthogonalTiledMap game_map: The game map this is the camera of
        """
        TiledMapCamera.__init__(self, game_map)

        self.viewport_x_offset = renpy.store.pink_otm_default_viewport_x_offset
        self.viewport_y_offset = renpy.store.pink_otm_default_viewport_y_offset
        self.viewport_width = renpy.store.pink_otm_default_viewport_width
        self.viewport_height = renpy.store.pink_otm_default_viewport_height

        self.camera_target = None
        self.xzoom = renpy.store.pink_otm_camera_default_zoom
        self.yzoom = renpy.store.pink_otm_camera_default_zoom

    def set_zoom(self, zoom):
        """
        Sets this camera's x and y zoom to the given value
        :param float zoom: the given zoom value
        """
        self.set_x_zoom(zoom)
        self.set_y_zoom(zoom)

    def set_x_zoom(self, xzoom):
        """
        Sets this camera's x zoom to the given value
        :param float xzoom: the given zoom value
        """
        self.x.range = renpy.store.pink_otm_current_map.image_size.x * self.xzoom
        self.xzoom = xzoom
        renpy.store.pink_otm_zoom_controller.original_xzoom = xzoom
        renpy.store.pink_otm_zoom_controller.target_xzoom = xzoom
        self.center_on_target()

    def set_y_zoom(self, yzoom):
        """
        Sets this camera's u zoom to the given value
        :param float yzoom: the given zoom value
        """
        self.y.range = renpy.store.pink_otm_current_map.image_size.y * self.yzoom
        self.yzoom = yzoom
        renpy.store.pink_otm_zoom_controller.original_yzoom = yzoom
        renpy.store.pink_otm_zoom_controller.target_yzoom = yzoom
        self.center_on_target()

    def camera_map_refresh(self):
        """
        Called by the pink_otm_call_map label, refreshing the camera for the new map.
        """
        self.x.range = renpy.store.pink_otm_current_map.image_size.x * self.xzoom
        self.y.range = renpy.store.pink_otm_current_map.image_size.y * self.yzoom

        self.center_on_target()

    @staticmethod
    def camera_target_for_object(target_object):
        """
        Which particular to pixel to center for a camera aimed at the given object.
        :param OrthogonalTiledMapGameObjectMobile target_object: the given target
        """
        centered_x = (
                target_object.x + target_object.camera_offset_x + int(target_object.width / 2))
        centered_y = (
                target_object.y + target_object.camera_offset_y + int(target_object.height / 2))

        return centered_x, centered_y

    def center_on_target(self):
        """
        Makes the camera center on its current target.
        """
        if self.camera_target is None:
            return

        centered_x, centered_y = self.camera_target_for_object(self.camera_target)
        self.center_on_location(centered_x, centered_y)

    def center_on_location(self, target_x, target_y):
        """
        Used the center the camera on a specific location.
        :param int target_x: x (in pixels) of the target location
        :param int target_y: y (in pixels) of the target location
        :return:
        """
        pink_otm_current_map = renpy.store.pink_otm_current_map  # type: OrthogonalTiledMap

        target_x *= self.xzoom
        target_y *= self.yzoom

        if (pink_otm_current_map.image_size.x * self.xzoom) >= self.viewport_width:
            max_x = (pink_otm_current_map.image_size.x * self.xzoom) - self.viewport_width
            centered_x = target_x - (self.viewport_width / 2)
            final_x = max(min(max_x, centered_x), 0)
        else:
            final_x = 0

        if (pink_otm_current_map.image_size.y * self.yzoom) >= self.viewport_height:
            max_y = (pink_otm_current_map.image_size.y * self.yzoom) - self.viewport_height
            centered_y = target_y - (self.viewport_height / 2)
            final_y = max(min(max_y, centered_y), 0)
        else:
            final_y = 0

        self.x.change(final_x)
        self.y.change(final_y)

    def smooth_zoom(self, target_zoom, zoom_speed):
        """
        Smoothly zooms out to the given value in the given number of seconds.
        :param float target_zoom: Which zoom level you want to zoom to (in ratio of the default)
        :param float zoom_speed: Amount of time to spend zooming (in seconds)
        """
        zoom_controller = renpy.store.pink_otm_zoom_controller  # Py 2
        if (
                zoom_controller.target_xzoom != target_zoom or
                zoom_controller.target_yzoom != target_zoom or
                zoom_controller.zoom_speed != zoom_speed):
            zoom_controller.start_zoom_st = zoom_controller.last_st
            zoom_controller.original_xzoom = self.xzoom
            zoom_controller.original_yzoom = self.yzoom
            zoom_controller.target_xzoom = target_zoom
            zoom_controller.target_yzoom = target_zoom
            zoom_controller.zoom_speed = zoom_speed

    def set_target(self, target):
        """
        Makes this object follow the given target, without immediately centering.
        :param OrthogonalTiledMapGameObjectMobile target: the given target
        """
        self.camera_target = target

    def switch_target(self, new_target):
        """
        Makes this object follow the given target, with immediate centering.
        :param OrthogonalTiledMapGameObject new_target: The given target.
        """
        self.camera_target = new_target
        self.center_on_target()

    def recenter_on_player(self):
        """
        Recenters the camera on the pc.
        """
        self.switch_target(renpy.store.pink_otm_current_pc)

    def spawn_camera_drone(self):
        """
        Spawns a camera drone and switches the camera to it. The camera drone is a pink collection object, which can
        then be given any kind of movement command.
        """
        drone_coord = self.camera_target.central_coord
        current_center_x, current_center_y = self.camera_target_for_object(self.camera_target)

        pink_otm_current_map = renpy.store.pink_otm_current_map  # type: OrthogonalTiledMap
        pink_otm_current_map.add_object_at_coord(
            renpy.store.pink_otm_camera_drone_init_dict, x_coord=drone_coord.x, y_coord=drone_coord.y)

        camera_drone = renpy.store.pink_otm_camera_drone
        drone_center_x, drone_center_y = self.camera_target_for_object(camera_drone)
        camera_drone.camera_offset_x = current_center_x - drone_center_x + camera_drone.camera_offset_x
        camera_drone.camera_offset_y = current_center_y - drone_center_y + camera_drone.camera_offset_y
        self.switch_target(renpy.store.pink_otm_camera_drone)


class OrthogonalTiledMap(TiledMapGame):
    npc_layer_dict = {
        "draworder": "topdown", "id": 999999999999, "name": "player_layer",
        "objects": [],
        "opacity": 1,
        "properties": [
            {"name": "layer_type", "type": "string", "value": "dynamic"},
            {"name": "z_order", "type": "int", "value": 100}],
        "type": "objectgroup", "visible": True, "x": 0, "y": 0}
    follower_layer_dict = {
        "draworder": "topdown", "id": 999999999999, "name": "follower_layer",
        "objects": [],
        "opacity": 1,
        "properties": [
            {"name": "layer_type", "type": "string", "value": "dynamic"},
            {"name": "z_order", "type": "int", "value": 99}],
        "type": "objectgroup", "visible": True, "x": 0, "y": 0}
    default_camera_variable = "pink_otm_current_camera"

    object_sprite_collection_class = OrthogonalTiledMapGameObjectSpriteCollection
    object_class = OrthogonalTiledMapGameObject
    tile_class = OrthogonalTiledMapGameTile
    tile_animated_class = OrthogonalTiledMapGameTileAnimated

    camera_class = OrthogonalTiledCamera

    def __init__(self, map_dict, path, **kwargs):
        """
        A single Orthogonal Tiled Map
        :param dict map_dict: The initialization dictionary for this map
        :param str path: The path at which this map can be found.
        """
        self.mobile_objects = []
        self.base_grid = OrthogonalTiledMapTiledBases(self, map_dict.get('width'), map_dict.get('height'))

        TiledMapGame.__init__(self, map_dict, path, **kwargs)

        # noinspection PyTypeChecker
        self.player_object = None  # type: OrthogonalTiledMapGameObjectSpriteCollection
        self.npc_layer = self.add_layer(self.npc_layer_dict)
        self.follower_layer = self.add_layer(self.follower_layer_dict)

        # Dictionaries for matching various types of input to functions.
        self.event_type_functions = renpy.store.pink_otm_event_type_functions  # type: dict
        self.key_down_functions = renpy.store.pink_otm_key_down_functions  # type: dict
        self.key_up_functions = renpy.store.pink_otm_key_up_functions  # type: dict
        self.key_held_functions = renpy.store.pink_otm_key_held_functions  # type: dict
        self.movement_keys = renpy.store.pink_otm_movement_keys  # type: dict
        self.joy_button_down_functions = renpy.store.pink_otm_joy_button_down_functions  # type: dict
        self.joy_button_up_functions = renpy.store.pink_otm_joy_button_up_functions  # type: dict
        self.mouse_button_down_functions = renpy.store.pink_otm_mouse_button_down_functions  # type: dict
        self.mouse_button_up_functions = renpy.store.pink_otm_mouse_button_up_functions  # type: dict

        # Variables related to player movement
        self._pc_movement_direction = None
        self._pc_is_running = False
        self._pc_walk_animation = renpy.store.pink_otm_default_walk_animation_name  # type: str
        self._pc_run_animation = renpy.store.pink_otm_default_run_animation_name  # type: str
        self._pc_walk_speed = renpy.store.pink_otm_current_walk_speed  # type: float
        self._pc_run_speed = renpy.store.pink_otm_current_run_speed  # type: float

        # Variables related to freezing the screen when the menu is open
        self._screen_freeze = False
        self._screen_freeze_gt = 0.0

        # Force check touch on go_to_map
        self.force_check_touch = False

        # Start up parallel processes
        self.parallel_processes = self.init_parallel_processes()
        for parallel_process in self.parallel_processes:
            parallel_process.on_init()

    def reload_keys(self):
        """
        Reloads the input keys this map responds to, processing any changes.
        """
        self.event_type_functions = renpy.store.pink_otm_event_type_functions  # type: dict
        self.key_down_functions = renpy.store.pink_otm_key_down_functions  # type: dict
        self.key_up_functions = renpy.store.pink_otm_key_up_functions  # type: dict
        self.key_held_functions = renpy.store.pink_otm_key_held_functions  # type: dict
        self.movement_keys = renpy.store.pink_otm_movement_keys  # type: dict
        self.joy_button_down_functions = renpy.store.pink_otm_joy_button_down_functions  # type: dict
        self.joy_button_up_functions = renpy.store.pink_otm_joy_button_up_functions  # type: dict
        self.mouse_button_down_functions = renpy.store.pink_otm_mouse_button_down_functions  # type: dict
        self.mouse_button_up_functions = renpy.store.pink_otm_mouse_button_up_functions  # type: dict

    def game_tick(self, gt):
        """
        Core gameplay loop, advancing the game by a single tick per run.
        :param float gt: The current game time in seconds
        """
        # Checks for a touch event on map start.
        if (
                self.force_check_touch and
                self.player_object is not None and
                renpy.store.pink_otm_current_event_name is None):
            self.force_check_touch = False
            self.check_touch(self.player_object)

        # If the gt has been reset for whatever reason, this updates the animation frame times for all animated tiles
        # to correspond to the reset st.
        if gt < self.last_gt:
            for animated_tile in self.animated_tiles:
                animated_tile.st_reset_frames(gt)

        # Checks whether a menu screen is open
        self.check_freeze_screen(gt)

        # Moves mobile objects
        self.check_player_movement()
        for mobile_object in self.mobile_objects:
            mobile_object.per_tick(gt)

        # Advances animations and checks conditional objects
        TiledMapGame.game_tick(self, gt)

        # Runs every parallel process once.
        for parallel_process in renpy.store.pink_otm_global_parallel_processes:
            parallel_process.per_tick(gt)
        for parallel_process in self.parallel_processes:
            parallel_process.per_tick(gt)

        renpy.exports.retain_after_load()

    def check_freeze_screen(self, gt):
        """
        Freezes and unfreezes the game depending on whether or not there are any menu screens open and whether the
        game is currently frozen because of that.
        :param float gt: The current game time in seconds
        """
        # Freezes the game when a menu screen is up.
        freeze_screen_open = self.freeze_screen_open()
        if not self._screen_freeze:
            # The game is currently not frozen.
            if freeze_screen_open:
                # A menu screen is open, so the screen should be frozen
                self._screen_freeze = True
                self._screen_freeze_gt = gt

                self.disable_controls()
                self.stop_running()
                self.freeze_map()
        else:
            # The game is currently frozen because of a menu screen
            if freeze_screen_open:
                # A menu screen is still open, so the game should remain frozen.
                return

            self._screen_freeze = False

            if gt == 0.0:
                # The game was just loaded from a save file.
                if renpy.store.pink_otm_current_event_type == 'static':
                    # However, the game should remain frozen because there is a static event going on. The movement GT
                    # should increment with self.last_gt so that the unfreezing process at the end
                    # of the event has the correct game time.
                    self._increment_all_gt(self.last_gt)
                else:
                    # No static event is going on, so controls are enabled, and the map is unfrozen without a GT
                    # increment
                    self.unfreeze_map(increment_gt=False)
                    if renpy.store.pink_otm_current_event_type is None:
                        self.check_held_keys()
                        self.enable_controls()
            else:
                # A menu screen was just closed.

                if renpy.store.pink_otm_current_event_type == 'static':
                    # However, the game should remain frozen because there is a static event going on. Because a
                    # static event ends in a GT reset, no movement/animation GT increment is required.
                    pass
                else:
                    # No static event is going on, so controls are enabled. The movement GT and animation GT should
                    # increment by the amount of time the menu was open.
                    gt_diff = gt - self._screen_freeze_gt

                    self._increment_all_gt(gt_diff)
                    self.unfreeze_map(increment_gt=False)
                    if renpy.store.pink_otm_current_event_type is None:
                        self.check_held_keys()
                        self.enable_controls()

    def _increment_all_gt(self, gt_diff):
        """
        Increments all animation and movement gts by the given amount, used if the game skips over a period of game time
        :param float gt_diff: The amount by which to increment gt.
        """
        self._increment_all_animation_gt(gt_diff)
        self._increment_all_movement_gt(gt_diff)

    def _increment_all_animation_gt(self, gt_diff):
        """
        Increments all animation gts by the given amount, used if the game skips over a period of game time
        :param float gt_diff: The amount by which to increment gt.
        """
        for animated_tile in self.animated_tiles:
            animated_tile.increment_gt(gt_diff)

    def _increment_all_movement_gt(self, gt_diff):
        """
        Increments all movement gts by the given amount, used if the game skips over a period of game time
        :param float gt_diff: The amount by which to increment gt.
        """
        for mobile_object in self.mobile_objects:
            mobile_object.increment_gt(gt_diff)

    @staticmethod
    def freeze_screen_open():
        """
        Determines whether or not a screen that the OTM should freeze on is currently open. The list of screen
        tag/layer combinations is contained in the pink_freeze_game_screens ren'py variable, which is set in the pink
        config
        """
        pink_freeze_game_screens = renpy.store.pink_freeze_game_screens  # Py2
        for tag, layer in pink_freeze_game_screens:
            if renpy.exports.get_screen(tag, layer=layer):
                return True
        return False

    def change_walk_speed(self, new_speed):
        """
        Changes the player walking speed to the given value.
        :param float new_speed: The player's new walking speed.
        """
        renpy.store.pink_otm_current_walk_speed = new_speed
        self._pc_walk_speed = new_speed

    def change_run_speed(self, new_speed):
        """
        Changes the player running speed to the given value.
        :param float new_speed: The player's new running speed.
        """
        renpy.store.pink_otm_current_run_speed = new_speed
        self._pc_run_speed = new_speed

    def check_player_movement(self):
        """
        The player's per-tick movement calculation
        """
        if (
                self.controls_enabled and
                self._pc_movement_direction is not None and
                len(self.player_object.control_stack) == 0
        ):
            # Movement animation and speed  # TODO not efficient?
            if self._pc_is_running:
                if self.player_object.move_animation != self._pc_run_animation:
                    self.player_object.control_stack.append(
                        ControlCommand('change_move_animation', animation_name=self._pc_run_animation)
                    )
                if self.player_object.movement_speed != self._pc_run_speed:
                    self.player_object.control_stack.append(
                        ControlCommand('change_movement_speed', quantity=self._pc_run_speed))
            else:
                if self.player_object.move_animation != self._pc_walk_animation:
                    self.player_object.control_stack.append(
                        ControlCommand('change_move_animation', animation_name=self._pc_walk_animation))
                if self.player_object.movement_speed != self._pc_walk_speed:
                    self.player_object.control_stack.append(
                        ControlCommand('change_movement_speed', quantity=self._pc_walk_speed))

            # Actual movement
            self.player_object.control_stack.append(
                ControlCommand(
                    "go_" + self._pc_movement_direction,
                    turn_and_move=renpy.store.pink_otm_turn_and_move,
                    animate_invalid_move=renpy.store.pink_otm_animate_invalid_move,
                    pop_invalid_move=True))

    def add_element(self, x, y, layer, game_object):
        """
        Adds a new element to the game map, and returns its displayable id. Checks for conditionality first.
        :param int x: Distance in pixels between the object's left side and the left border of the map.
        :param int y: Distance in pixels between the object's top side and the top border of the map.
        :param TiledMapGameLayer layer: The layer the object is a part of.
        :param OrthogonalTiledMapGameTile|OrthogonalTiledMapGameTileAnimated|OrthogonalTiledMapGameObject game_object:
        The object to add
        :rtype: int
        """
        TiledMapGame.add_element(self, x, y, layer, game_object)
        if game_object.is_conditional:
            if game_object.condition_met():
                self._add_element(x, y, layer, game_object)
        else:
            self._add_element(x, y, layer, game_object)

    def _add_element(self, x, y, layer, game_object):
        """
        Subfunction of add_element
        :param int x: Distance in pixels between the object's left side and the left border of the map.
        :param int y: Distance in pixels between the object's top side and the top border of the map.
        :param TiledMapGameLayer layer: The layer the object is a part of.
        :param OrthogonalTiledMapGameTile|OrthogonalTiledMapGameTileAnimated|OrthogonalTiledMapGameObject game_object:
        The object to add
        :rtype: int
        """
        if isinstance(game_object, OrthogonalTiledMapGameObjectMobile):
            self.mobile_objects.append(game_object)

        # Override coords if the object has its own. This accommodates consistent objects.
        if hasattr(game_object, 'x'):
            x = game_object.x
        if hasattr(game_object, 'y'):
            y = game_object.y
        self.base_grid.add_element(x, y, layer, game_object)

        if isinstance(game_object, OrthogonalTiledMapGameObjectSpriteCollection):
            # If adding post-initiation, rebases frames
            if self.last_gt != 0.0:
                game_object.st_reset_frames(self.last_gt)

            # Auto-centers
            if game_object.should_be_auto_centered():
                central_coord = game_object.central_coord
                game_object.set_to_coords(central_coord.x, central_coord.y, orientation=game_object.start_orientation)

            if renpy.store.pink_otm_current_event_type == 'static':
                game_object.freeze()
            elif renpy.store.pink_otm_current_event_type == "dynamic":
                game_object.save()

    def remove_element(self, element):
        """
        Removes the given element from the map
        :param OrthogonalTiledMapGameTile|OrthogonalTiledMapGameTileAnimated|OrthogonalTiledMapGameObject element:
        The element to remove
        """
        if isinstance(element, OrthogonalTiledMapGameObjectMobile):
            self.mobile_objects.remove(element)

        element_id = element.displayable_id
        self.base_grid.remove_element(element_id)
        TiledMapGame.remove_element(self, element)

    def add_object_at_coord(
            self, object_dict, x_coord=0, y_coord=0, is_player=False, orientation=None, layer='npc'
    ):
        """
        Adds a new element to the game map at the given coordinates.
        :param dict object_dict: The initialization dictionary for the object
        :param int x_coord: The x coordinate which to center the object on
        :param int y_coord: The y coordinate which to center the object on
        :param bool is_player: Whether this object is the player character
        :param str orientation: The orientation to which to set the new object.
        :param str layer: Either 'npc' or 'follower'
        """
        if layer == "npc":
            target_layer = self.npc_layer
        else:
            target_layer = self.follower_layer
        new_object = self.add_object(object_dict=object_dict, x=0, y=0, layer=target_layer)
        new_object.set_to_coords(x_coord, y_coord, orientation=orientation)

        if is_player:
            self.camera.set_target(new_object)
            self.player_object = new_object

        return new_object

    def pixel_to_coord(self, pixel_loc, is_x=True):
        """
        Returns the coordinates of the given pixel on the current map.
        :param int pixel_loc: The given pixel location.
        :param bool is_x: Whether the given pixel location is an x (if True), or a y (if False)
        :return: The coordinates of the given pixel on the current map.
        :rtype: int
        """
        if is_x:
            return int(pixel_loc / self.tile_size.x)
        else:
            return int(pixel_loc / self.tile_size.y)

    def check_activation_code(self):
        """
        Called by the interaction key, this checks whether there is code to be interacted with in the current
        location. If so, this function also executes that code
        """
        if self.player_object is None:
            return

        player_base = self.base_grid.map_elements[self.player_object.displayable_id]

        for x, y in player_base.coords:
            if self.player_object.orientation == "up":
                code, code_cause = self.base_grid.get_activation_code(x, y - 1, ignore_element=player_base)
            elif self.player_object.orientation == "right":
                code, code_cause = self.base_grid.get_activation_code(x + 1, y, ignore_element=player_base)
            elif self.player_object.orientation == "down":
                code, code_cause = self.base_grid.get_activation_code(x, y + 1, ignore_element=player_base)
            elif self.player_object.orientation == "left":
                code, code_cause = self.base_grid.get_activation_code(x - 1, y, ignore_element=player_base)
            else:
                break

            """Executes code"""
            if code:
                exec(code)

    def check_activation_event(self):
        """
        Called by the interaction key, this checks whether there is an event to be interacted with in the current
        location. If so, this function also calls that event.
        """
        if self.player_object is None:
            return

        player_base = self.base_grid.map_elements[self.player_object.displayable_id]

        for x, y in player_base.coords:
            if self.player_object.orientation == "up":
                event, event_cause = self.base_grid.get_activation_event(x, y - 1, ignore_element=player_base)
            elif self.player_object.orientation == "right":
                event, event_cause = self.base_grid.get_activation_event(x + 1, y, ignore_element=player_base)
            elif self.player_object.orientation == "down":
                event, event_cause = self.base_grid.get_activation_event(x, y + 1, ignore_element=player_base)
            elif self.player_object.orientation == "left":
                event, event_cause = self.base_grid.get_activation_event(x - 1, y, ignore_element=player_base)
            else:
                break

            """Jumps to event in question"""
            if event:
                event_args, event_kwargs = event_cause.event_on_activate_args, event_cause.event_on_activate_kwargs
                return self.trigger_event(event, event_cause, *event_args, **event_kwargs)

    def check_touch(self, triggering_object):
        """
        Checks whether the given object should trigger any touch code or events. If so, this function also executes that
        code and/or calls that event.
        :param OrthogonalTiledMapGameObjectBase triggering_object: The given object
        """
        code, code_cause, event, event_cause = None, None, None, None
        if renpy.store.pink_otm_current_event_type is None:
            if triggering_object is self.player_object:
                player_base = self.base_grid.map_elements[self.player_object.displayable_id]
                for x, y in player_base.coords:
                    code, code_cause = self.base_grid.get_touch_code(
                        x, y, ignore_element=player_base)
                    event, event_cause = self.base_grid.get_touch_event(
                        x, y, ignore_element=player_base)
            else:
                code, code_cause = triggering_object.code_on_touch, triggering_object
                event, event_cause = triggering_object.event_on_touch, triggering_object

        if code:
            exec(code)

        """Jumps to event in question"""
        if event:
            event_args, event_kwargs = event_cause.event_on_touch_args, event_cause.event_on_touch_kwargs
            return self.trigger_event(event, event_cause, *event_args, **event_kwargs)

    @staticmethod
    def trigger_event(event_label, event_cause, *args, **kwargs):
        """
        Triggers the event with the given label
        :param str event_label: The given event label
        :param OrthogonalTiledMapGameObjectBase event_cause: The object causing the event to be triggered
        """
        renpy.store.pink_otm_event_trigger = event_cause
        renpy.store.pink_otm_current_event_name = event_label
        renpy.exports.call(event_label, *args, **kwargs)

    def event_key_down(self, ev, x, y, st):  # noqa
        """
        Checks whether this should execute any functions on a given key_down event, and executes them if so.
        :param ev: The event
        :param x: event x
        :param y: event y
        :param st: event st
        """
        if ev.key in self.movement_keys:
            self._movement_key_down(ev, x, y, st)
        elif ev.key in self.key_down_functions:
            self.key_down_functions[ev.key](self, ev, x, y, st)  # noqa

    def event_key_up(self, ev, x, y, st):  # noqa
        """
        Checks whether this should execute any functions on a given key_up event, and executes them if so.
        :param ev: The event
        :param x: event x
        :param y: event y
        :param st: event st
        """
        if ev.key in self.movement_keys:
            self._movement_key_up(ev, x, y, st)
        if ev.key in self.key_up_functions:
            self.key_up_functions[ev.key](self, ev, x, y, st)  # noqa

    def event_joy_button_down(self, ev, x, y, st):
        """
        Checks whether this should execute any functions on a given joy_button_down event, and executes them if so.
        :param ev: The event
        :param x: event x
        :param y: event y
        :param st: event st
        """
        if ev.button in self.joy_button_down_functions:
            self.joy_button_down_functions[ev.button](self, ev, x, y, st)  # noqa

    def event_joy_button_up(self, ev, x, y, st):
        """
        Checks whether this should execute any functions on a given joy_button_up event, and executes them if so.
        :param ev: The event
        :param x: event x
        :param y: event y
        :param st: event st
        """
        if ev.button in self.joy_button_up_functions:
            self.joy_button_up_functions[ev.button](self, ev, x, y, st)  # noqa

    def event_mouse_button_down(self, ev, x, y, st):
        """
        Checks whether this should execute any functions on a given mouse_button_down event, and executes them if so.
        :param ev: The event
        :param x: event x
        :param y: event y
        :param st: event st
        """
        if ev.button in self.mouse_button_down_functions:
            self.mouse_button_down_functions[ev.button](self, ev, x, y, st)  # noqa

    def event_mouse_button_up(self, ev, x, y, st):
        """
        Checks whether this should execute any functions on a given mouse_button_up event, and executes them if so.
        :param ev: The event
        :param x: event x
        :param y: event y
        :param st: event st
        """
        if ev.button in self.mouse_button_up_functions:
            self.mouse_button_up_functions[ev.button](self, ev, x, y, st)  # noqa

    def event_joy_axis_motion(self, ev, x, y, st):  # noqa
        """
        Checks whether this should execute any functions on a given joy_axis_motion event, and executes them if so.
        :param ev: The event
        :param x: event x
        :param y: event y
        :param st: event st
        """
        axis = ev.axis
        value = ev.value

        # Is this a movement on an x axis or a y axis?
        if axis % 2 == 0:
            x_axis = True
        else:
            x_axis = False

        # If movement is set to 0, stops movement
        if value == 0:
            self.stop_movement_all()

        # Otherwise, moves in the right direction
        elif x_axis:
            if value < 0:
                self.start_movement("left")
            else:
                self.start_movement("right")
        else:
            if value < 0:
                self.start_movement("down")
            else:
                self.start_movement("up")
        raise renpy.exports.IgnoreEvent()

    def event_joy_ball_motion(self, ev, x, y, st):  # noqa
        """
        Checks whether this should execute any functions on a given joy_ball_motion event, and executes them if so.
        :param ev: The event
        :param x: event x
        :param y: event y
        :param st: event st
        """
        # This function is left as an exercise to the reader.
        pass

    def event_joy_hat_motion(self, ev, x, y, st):  # noqa
        """
        Checks whether this should execute any functions on a given joy_hat_motion event, and executes them if so.
        :param ev: The event
        :param x: event x
        :param y: event y
        :param st: event st
        """
        hat_x, hat_y = ev.value
        if hat_x == hat_y == 0:
            return
        if abs(hat_x) > abs(hat_y):
            if hat_x < 0:
                self.turn_player("left")
            else:
                self.turn_player("right")
        else:
            if hat_y < 0:
                self.turn_player("down")
            else:
                self.turn_player("up")
        raise renpy.exports.IgnoreEvent()

    def _movement_key_down(self, ev, x, y, st):  # noqa
        """
        Checks whether this should execute any functions on a given key_down event if said key is registered as
        a movement key, and executes them if so.
        :param ev: The event
        :param x: event x
        :param y: event y
        :param st: event st
        """
        direction = self.movement_keys[ev.key]
        if ev.mod & pygame.KMOD_ALT and renpy.store.pink_otm_alt_turning_enabled:
            self.turn_player(direction)
        else:
            self.start_movement(direction)
        raise renpy.exports.IgnoreEvent()

    def _movement_key_up(self, ev, x, y, st):  # noqa
        """
        Checks whether this should execute any functions on a given key_up event if said key is registered as
        a movement key, and executes them if so.
        :param ev: The event
        :param x: event x
        :param y: event y
        :param st: event st
        """
        direction = self.movement_keys[ev.key]
        self.stop_movement(direction)
        raise renpy.exports.IgnoreEvent()

    def check_held_keys(self):
        """
        Checks the currently held keys, and whether any functions should be executed depending on those keys. Used
        to start movement immediately on entering a new map or ending an event.
        """
        self.check_held_direction()
        keys = pygame.key.get_pressed()

        for held_key in self.key_held_functions:
            if keys[held_key]:
                self.key_held_functions[held_key](self)  # noqa

    @staticmethod
    def allowed_to_run():
        """
        :return: Whether the player is currently allowed to run
        :rtype: bool
        """
        return not (renpy.store.pink_otm_current_map.forbid_run or renpy.store.pink_otm_forbid_run)

    def forced_to_run(self):
        """
        :return: Whether the player is currently forced to run
        :rtype: bool
        """
        return self.allowed_to_run() and (renpy.store.pink_otm_current_map.force_run or renpy.store.pink_otm_force_run)

    def console_key_down(self, ev, x, y, st):  # noqa
        """
        Executes an event where the current console key is pressed. Checks whether the player is allowed to
        access the console, and whether the shift key is held before jumping to console label.
        """
        if (renpy.store.config.console or renpy.store.config.developer) and (ev.mod & pygame.KMOD_SHIFT):
            # This is a call rather than a jump, as the label makes use of leave_otm, which pops the
            # latest item off the call stack.
            renpy.exports.call("pink_otm_developer_start")
            raise renpy.exports.IgnoreEvent()

    def run_key_down(self, ev, x, y, st):  # noqa
        """
        Executes an event where the run key is pressed down. Checks whether the player is allowed to start running,
        and makes the player start running if so.
        """
        self.start_running()
        raise renpy.exports.IgnoreEvent()

    def run_key_up(self, ev, x, y, st):  # noqa
        """
        Executes an event where the run key is released. Checks whether the player is allowed to stop running,
        and makes the player stop running if so.
        """
        self.stop_running()
        raise renpy.exports.IgnoreEvent()

    def start_running(self):
        """
        Makes the player start running if allowed.
        """
        if self.allowed_to_run():
            self._pc_is_running = True

    def stop_running(self):
        """
        Makes the player stop running if allowed.
        """
        if not self.forced_to_run():
            self._pc_is_running = False

    def turn_player(self, direction):
        """
        Adds a turn command in the given direction to the player's control stack
        :param str direction: the given direction
        """
        self.player_object.control_stack.append(ControlCommand(command_type='turn_' + direction))

    def start_movement(self, direction):
        """
        Makes the player start movement in the given direction
        :param str direction: the given direction
        """
        self._pc_movement_direction = direction

    def stop_movement(self, direction):
        """
        Makes the player stop movement in the given direction
        :param str direction: the given direction
        """
        if self._pc_movement_direction == direction:
            if (
                    len(self.player_object.control_stack) > 0 and
                    self.player_object.control_stack[0].type == "go_" + direction
            ):
                self.player_object.control_stack.pop(0)
            self._pc_movement_direction = None
            self.check_held_direction()

    def check_held_direction(self):
        """
        Checks whether any movement keys are currently pressed, and starts movement keys in that direction if so.
        """
        keys = pygame.key.get_pressed()
        for movement_key, direction in self.movement_keys.items():
            if keys[movement_key]:
                self.start_movement(direction)

    def stop_movement_all(self):
        """
        Stops the player's movement entirely.
        """
        self.player_object.control_stack = []
        self._pc_movement_direction = None

    def player_interaction(self, ev, x, y, st):  # noqa
        """
        Executes an event where the interaction key is released. Checks whether there is any code or event to
        execute in the given location, and executes that code and/or event.
        """
        if renpy.store.pink_otm_current_event_type is None:
            self.check_activation_code()
            self.check_activation_event()
            raise renpy.exports.IgnoreEvent()

    def disable_controls(self):
        """
        Stops all movement and disables controls
        """
        self.stop_movement_all()
        self.controls_enabled = False

    def enable_controls(self):
        """
        Enables controls
        """
        self.controls_enabled = True

    def end_all_interactions(self):
        """
        Ends the interactions on all Mobile Objects on this map.
        """
        for mobile_object in self.mobile_objects:
            if mobile_object.currently_interacting:
                mobile_object.end_interaction()

    def freeze_map(self):
        """
        Freezes the map, making all objects stop dead in their tracks and all animations halt.
        """
        for mobile_object in self.mobile_objects:
            mobile_object.freeze()
        for animated_tile in self.animated_tiles:
            animated_tile.pause_animation()

    def unfreeze_map(self, increment_gt=True):
        """
        Unfreezes the map, making all objects continue where they stopped, and restarts animations
        :param bool increment_gt: If True, increments the gt of all mobile and animated objects to account for the
        amount of time they were frozen.
        """
        for mobile_object in self.mobile_objects:
            mobile_object.unfreeze(increment_gt=increment_gt)
        for animated_tile in self.animated_tiles:
            animated_tile.unpause_animation()

    def halt_map(self):
        """
        Saves the position of all objects on the map, and halts them, as well as blocking player controls. Intended
        to be used for labels with movements (such as cutscenes).
        """
        for mobile_object in self.mobile_objects:
            if mobile_object is not self.player_object:
                mobile_object.save_and_halt()

    def unhalt_map(self):
        """
        Restores all sprites to the location they were saved from.
        """
        for mobile_object in self.mobile_objects:
            if mobile_object is not self.player_object:
                mobile_object.load_and_continue()

    def halt_player(self):
        self.player_object.force_stand()
        self.stop_running()
        self.player_object.movement_speed = self._pc_walk_speed

        for follower in renpy.store.pink_otm_followers:
            follower.movement_speed = self._pc_walk_speed

    def init_parallel_processes(self):
        """
        :return: Initiates all parallel process classes used by this map, and returns a list of them
        :rtype: list
        """
        if 'parallel_processes' in self.properties:
            return eval(self.properties['parallel_processes'])
        else:
            return []

    @property
    def background_music(self):
        """
        :return: The background music track to play for this map, or whether to 'continue' the background music of the
        previous map or whether to 'stop' the currently playing background music.
        :rtype: str
        """
        if 'background_music' in self.properties:
            return self.properties['background_music']
        else:
            return "continue"

    @property
    def background_ambience(self):
        """
        :return: The background ambience track to play for this map, or whether to 'continue' the background ambience
        of the previous map or whether to 'stop' the currently playing background ambience.
        :rtype: str
        """
        if 'background_ambience' in self.properties:
            return self.properties['background_ambience']
        else:
            return "stop"

    @property
    def forbid_run(self):
        """
        :return: Whether running is forbidden on this map
        :rtype: bool
        """
        if 'forbid_run' in self.properties:
            return self.properties['forbid_run']
        else:
            return False

    @property
    def force_run(self):
        """
        :return: Whether running is forced on this map
        :rtype: bool
        """
        if 'force_run' in self.properties:
            return self.properties['force_run']
        else:
            return False


def restore_rollback_settings():
    """
    Restores the rollback settings from before the pink engine was initialized.
    """
    renpy.store.config.rollback_enabled = renpy.store.pink_otm_rollback_restore


def go_to_map(target_map, x_coord=0, y_coord=0, orientation=None, spawn_player=True):
    """
    Initializes the given OTM map and moves the game to said map
    :param str target_map: The path of the given OTM map
    :param int x_coord: The x coord at which to place the player character
    :param int y_coord: The y coord at which to place the player character
    :param str orientation: The orientation at which to place the player character
    :param bool spawn_player: Whether to add the player character to the map
    """
    renpy.store.pink_otm_current_event_name = None
    if renpy.store.pink_otm_current_event_type is not None:
        end_event(renpy.store.pink_otm_current_event_type, no_jump=True)
    elif renpy.store.pink_otm_current_map:  # To prevent popping the very first go_to_map
        # Pops
        renpy.exports.pop_call()

    # Rollback is restored, if enabled, but can't rollback to before start of event.
    restore_rollback_settings()
    renpy.exports.block_rollback()

    target_x_zoom = renpy.store.pink_otm_camera_default_zoom
    target_y_zoom = renpy.store.pink_otm_camera_default_zoom

    if renpy.store.pink_otm_current_map is not None:
        renpy.store.pink_otm_current_map.stop_predicting()
        if orientation is None and renpy.store.pink_otm_current_map.player_object is not None:
            orientation = renpy.store.pink_otm_current_map.player_object.orientation
        target_x_zoom = renpy.store.pink_otm_current_map.camera.xzoom
        target_y_zoom = renpy.store.pink_otm_current_map.camera.yzoom

    map_file = renpy.exports.file("pink_engine/orthogonal_tiled_maps/" + target_map)
    map_dict = json.load(map_file)
    new_map = OrthogonalTiledMap(map_dict, "pink_engine/orthogonal_tiled_maps/" + target_map)
    renpy.store.pink_otm_current_map = new_map

    new_map.camera.set_x_zoom(target_x_zoom)
    new_map.camera.set_y_zoom(target_y_zoom)

    if spawn_player:
        new_map.add_object_at_coord(
            renpy.store.pink_otm_pc_init_dict, x_coord, y_coord, is_player=True, orientation=orientation)
        renpy.store.pink_otm_current_pc = new_map.player_object

        new_map.check_held_keys()

        if new_map.force_run:
            new_map.start_running()
        if new_map.forbid_run:
            new_map.stop_running()

        # FOLLOWERS
        old_follower_objects = renpy.store.pink_otm_followers[:]  # py2
        renpy.store.pink_otm_followers = []
        for old_follower_object in old_follower_objects:
            add_player_follower(old_follower_object.sprite_collection_path)
        for follower in renpy.store.pink_otm_followers:
            follower.movement_speed = new_map.player_object.movement_speed
            follower.move_animation = new_map.player_object.move_animation

    else:
        renpy.store.pink_otm_current_pc = None

    # MUSIC
    if new_map.background_music == "stop":
        renpy.exports.music.stop()
    elif new_map.background_music != "continue":
        renpy.exports.music.play(new_map.background_music, loop=True, if_changed=True)

    # AMBIENCE
    if new_map.background_ambience == "stop":
        renpy.exports.music.stop(channel='sound')
    elif new_map.background_ambience != "continue":
        renpy.exports.music.play(new_map.background_ambience, channel='sound', loop=True, if_changed=True)

    new_map.force_check_touch = True

    renpy.exports.jump('pink_otm_call_map')


def force_instant_transition_on_event_end():
    """
    Temporarily overrides config.window_hide_transition so as to make the pink engine not have any window hide
    transition during the current event. Necessary if you have an event that has dialogue and ends on a go_to_map
    statement, and you want that go_to_map to be performed instantly.
    """
    renpy.store.pink_otm_window_hide_cache = renpy.store.config.window_hide_transition
    renpy.store.config.window_hide_transition = None


def start_event(event_type, interaction_reaction=True):
    """
    Starts a pink engine event with the given type
    :param str event_type: The given event type (static, dynamic or continuous)
    :param bool interaction_reaction: Whether to play the default interaction reaction
    """
    # The pink engine does not expect a return statement when it calls a label, it merely uses calls to pass along
    # variables
    renpy.exports.pop_call()

    renpy.store.pink_otm_current_event_type = event_type
    renpy.store.pink_otm_current_map.disable_controls()
    renpy.store.pink_otm_event_interaction_reaction = False

    # Rollback is restored, if enabled, but can't rollback to before start of event.
    restore_rollback_settings()
    renpy.exports.block_rollback()

    # Makes player stand still, as well as stopping their running
    renpy.store.pink_otm_current_map.halt_player()

    # Event type-dependent map stopping.
    if event_type == "static":
        renpy.store.pink_otm_current_map.freeze_map()
    elif event_type == "dynamic":
        renpy.store.pink_otm_current_map.halt_map()
    elif event_type == "continuing":
        pass

    # Default interaction
    if interaction_reaction:
        if event_type == "dynamic" and hasattr(renpy.store.pink_otm_event_trigger, 'default_interaction'):
            renpy.store.pink_otm_event_interaction_reaction = True
            renpy.store.pink_otm_event_trigger.default_interaction()
        elif event_type == "continuing" and hasattr(renpy.store.pink_otm_event_trigger, 'save_and_default_interaction'):
            renpy.store.pink_otm_event_interaction_reaction = True
            renpy.store.pink_otm_event_trigger.save_and_default_interaction()

    # Show screen pink_otm_map
    renpy.exports.show_screen('pink_otm_map')


def end_event(event_type, no_jump=False):
    """
    Ends a pink engine event with the given type
    :param str event_type: The given event type (static, dynamic or continuous)
    :param bool no_jump: Whether to jump to the otm_call_map label at the end of this event. Should be False if the
    event ended with going to a different map.
    """
    # In case any objects got removed right before the end of the event
    renpy.store.pink_otm_current_map.check_conditional_objects()

    if event_type == "static":
        renpy.store.pink_otm_current_map.unfreeze_map()
    elif event_type == "dynamic":
        renpy.store.pink_otm_current_map.unhalt_map()
    elif event_type == "continuing" and renpy.store.pink_otm_event_interaction_reaction is True:
        renpy.store.pink_otm_event_trigger.load_and_continue()

    renpy.store.pink_otm_event_trigger_object = None
    renpy.store.pink_otm_current_event_type = None
    renpy.store.pink_otm_current_event_name = None
    renpy.store.pink_otm_event_interaction_reaction = None

    renpy.store.pink_otm_current_map.enable_controls()
    renpy.store.pink_otm_current_map.check_held_keys()

    if renpy.store.pink_otm_current_map.force_run:
        renpy.store.pink_otm_current_map.start_running()

    # renpy.exports.scene()
    if not no_jump:
        renpy.exports.jump('pink_otm_call_map')


def start_static_event(interaction_reaction=True):
    """
    Starts a static event
    :param bool interaction_reaction: Whether to play the default interaction reaction
    """
    start_event('static', interaction_reaction)


def start_dynamic_event(interaction_reaction=True):
    """
    Starts a static event
    :param bool interaction_reaction: Whether to play the default interaction reaction
    """
    start_event('dynamic', interaction_reaction)


def start_continuing_event(interaction_reaction=True):
    """
    Starts a static event
    :param bool interaction_reaction: Whether to play the default interaction reaction
    """
    start_event('continuing', interaction_reaction)


def end_current_event():
    """
    Ends the current event
    """
    end_event(renpy.store.pink_otm_current_event_type)


def end_static_event():  # here for the sake of compatibility.
    """
    Ends the current event
    """
    end_current_event()


def end_dynamic_event():  # here for the sake of compatibility.
    """
    Ends the current event
    """
    end_current_event()


def end_continuing_event():  # there for the sake of consistency
    """
    Ends the current event
    """
    end_current_event()


def leave_otm():
    """
    Starts an event without an otm background.
    """
    # Pops the call that the pink engine made.
    renpy.exports.pop_call()

    renpy.store.pink_otm_player_coords = renpy.store.pink_otm_current_map.player_object.central_coord
    renpy.store.pink_otm_current_map.stop_predicting()
    renpy.store.pink_otm_map_name = os.path.split(renpy.store.pink_otm_current_map.path)[1]
    renpy.store.pink_otm_current_camera = None
    restore_rollback_settings()
    renpy.exports.block_rollback()

    # Restore autosave_on_choice configuration
    renpy.store.config.autosave_on_choice = renpy.store.pink_otm_autosave_on_choice_cache
    renpy.store.pink_otm_autosave_on_choice_cache = None

    renpy.exports.hide_screen('pink_otm_map')


def return_to_otm():
    """
    Returns to the orthogonal tiled map that was last active.
    """
    renpy.exports.jump('pink_otm_call_map')


def in_place(target_object_name, x_coord=None, y_coord=None, orientation=None, finished_moving=True):
    """
    Returns True if the character is in the given location.
    :param target_object_name: The ren'py variable for the given target object
    :param int x_coord: the given x coordinate. None for wildcard.
    :param int y_coord: the given y coordinate. None for wildcard.
    :param str orientation: The given orientation. None for wildcard.
    :param bool finished_moving: Whether or not the character must have completed their movement.
    :return:
    """
    target_object = getattr(renpy.store, target_object_name)

    if target_object is None or target_object.displayable_id is None:
        return False
    else:
        return target_object.in_place(
            x_coord=x_coord, y_coord=y_coord, orientation=orientation, finished_moving=finished_moving)


# Follower functions
def add_player_follower(sprite_collection_path):
    """
    :param str sprite_collection_path: The name of the sprite collection the follower should use.
    """
    follower_list = renpy.store.pink_otm_followers
    if len(follower_list) == 0:
        target_str = '"pink_otm_current_pc"'
        target = renpy.store.pink_otm_current_pc
    else:
        target_str = '"pink_otm_followers[' + str(len(follower_list) - 1) + ']"'
        target = follower_list[len(follower_list) - 1]

    follower_dict = {
        'properties': [
            {"name": "sprite_collection_path", "type": "string", "value": sprite_collection_path},
            {"name": "can_always_move", "type": "bool", "value": True},
            {
                "name": "default_control_stack",
                "type": "string",
                "value": "[\nControlCommand(\"go_to\", "
                         "animate_invalid_move=False, "
                         "target=" + target_str + ", "
                         "prev_coord=True)]",
            },
            {"name": "move_to", "type": "string", "value": None},
            {"name": "move_from", "type": "string", "value": None}  # TODO speed
        ]}
    target_coord = target.central_coord
    follower_object = renpy.store.pink_otm_current_map.add_object_at_coord(
        follower_dict, target_coord.x, target_coord.y, orientation=renpy.store.pink_otm_current_pc.orientation,
        layer="follower")
    renpy.store.pink_otm_followers.append(follower_object)


def remove_all_followers():
    for follower_object in renpy.store.pink_otm_followers:
        renpy.store.pink_otm_current_map.remove_element(follower_object)
    renpy.store.pink_otm_followers = []


def remove_follower_by_index(index):
    if index > len(renpy.store.pink_otm_followers) - 1:
        return

    renpy.store.pink_otm_current_map.remove_element(renpy.store.pink_otm_followers[index])
    renpy.store.pink_otm_followers.pop(index)
    _recalculate_follower_movement()


def remove_follower_by_path(sprite_collection_path, remove_all=False):
    removal_indices = []
    for x, follower in enumerate(renpy.store.pink_otm_followers):
        if follower.sprite_collection_path == sprite_collection_path:
            removal_indices.append(x)
            if not remove_all:
                break

    for index in sorted(removal_indices, reverse=True):
        renpy.store.pink_otm_current_map.remove_element(renpy.store.pink_otm_followers[index])
        renpy.store.pink_otm_followers.pop(index)

    _recalculate_follower_movement()


def _recalculate_follower_movement():
    for x, follower in enumerate(renpy.store.pink_otm_followers):
        if x == 0:
            target_str = "pink_otm_current_pc"
        else:
            target_str = "pink_otm_followers[" + str(x - 1) + "]"

        follower.control_stack = [
            ControlCommand(
                "go_to", animate_invalid_move=False, target=target_str, prev_coord=True)]


# Developer functions
def get_all_map_filenames():
    """
    :return: A list of all OTM map filenames
    :rtype: list
    """
    map_filenames = []

    all_renpy_filenames = renpy.exports.list_files()
    for renpy_filename in all_renpy_filenames:  # type: str
        if renpy_filename.startswith("pink_engine/orthogonal_tiled_maps/") and renpy_filename[-5:] == ".json":
            map_filenames.append(renpy_filename[34:])

    return map_filenames


def get_all_tileset_filenames():
    """
    :return: A list of all tileset filenames
    :rtype: list
    """
    tileset_filenames = []

    all_renpy_filenames = renpy.exports.list_files()
    for renpy_filename in all_renpy_filenames:  # type: str
        if renpy_filename.startswith("pink_engine/tilesets/") and renpy_filename[-5:] == ".json":
            tileset_filenames.append(renpy_filename[21:])

    return tileset_filenames


def get_all_sprite_filenames():
    """
    :return: A list of all sprite collection filenames
    :rtype: list
    """
    sprite_filenames = []

    all_renpy_filenames = renpy.exports.list_files()
    for renpy_filename in all_renpy_filenames:  # type: str
        if renpy_filename.startswith("pink_engine/sprite_collections/") and renpy_filename[-5:] == ".json":
            sprite_filenames.append(renpy_filename[31:])

    return sprite_filenames


def get_map_data(filename):
    """
    :param str filename: The given filename
    :return: A dictionary containing the properties of the otm map at the given filename
    :rtype: dict
    """
    json_file = renpy.exports.file("pink_engine/orthogonal_tiled_maps/" + filename)
    map_dict = json.load(json_file)

    grid_size = str(map_dict.get('width')) + "x" + str(map_dict.get('height'))
    tile_size = str(map_dict.get('tilewidth')) + "x" + str(map_dict.get('tileheight'))
    image_size = (str(map_dict.get('width') * map_dict.get('tilewidth')) +
                  "x" + str(map_dict.get('height') * map_dict.get('tileheight')))
    layer_count = str(len(map_dict.get('layers')))
    tiled_version = map_dict.get('tiledversion')

    properties = []
    if map_dict.get('properties') is not None:
        for map_property in map_dict.get('properties'):
            properties.append(map_property['name'])

    tilesets = []
    if map_dict.get('tilesets') is not None:
        for tileset_dict in map_dict.get('tilesets'):
            tilesets.append(os.path.basename(tileset_dict['source']))

    return grid_size, tile_size, image_size, layer_count, tiled_version, properties, tilesets


def get_tileset_data(filename):
    """
    :param str filename: The given filename
    :return: A dictionary containing the properties of the tileset at the given filename
    :rtype: dict
    """
    json_file = renpy.exports.file("pink_engine/tilesets/" + filename)
    tileset_dict = json.load(json_file)

    tile_count = str(len(tileset_dict.get('tiles')))
    tiled_version = tileset_dict.get('tiledversion')

    return tile_count, tiled_version


def get_sprite_data(filename):
    """
    :param str filename: The given filename
    :return: A dictionary containing the properties of the sprite collection at the given filename
    :rtype: dict
    """
    json_file = renpy.exports.file("pink_engine/sprite_collections/" + filename)
    sprite_dict = json.load(json_file)

    animation_names = set()
    animations = []

    for animation in sprite_dict['animations']:
        animation_names.add(animation['name'])

        frame_count = len(animation['images'])
        if frame_count > 1:
            ani_length = frame_count * animation['default_speed']
            animations.append(animation['name'] + " [" + str(frame_count) + " frames, " + str(ani_length) + " seconds]")
        else:
            animations.append(animation['name'] + " [static image]")

    if (
            "stand_down" in animation_names and "stand_up" in animation_names and
            "stand_left" in animation_names and "stand_right" in animation_names and
            "walk_down" in animation_names and "walk_up" in animation_names and
            "walk_left" in animation_names and "walk_right" in animation_names and
            "run_down" in animation_names and "run_up" in animation_names and
            "run_left" in animation_names and "run_right" in animation_names
    ):
        pc_sprite = True
    else:
        pc_sprite = False

    return animations, pc_sprite


# New feature: followers
# TODO global var for PC follower game objects. Used to spawn new followers on new map, and give orders to followers.
#  New functions: halt_followers, unspawn_followers, add_follower, remove_follower
#  some way to make characters invisible without fully removing them? So that followers can hide during a cutscene,
#  but are in logical positions after the cutscene ends and they reappear?
#
# TODO Autocenter fix for large NPC
# TODO door stuff
# TODO run into a room fix


# New feature: moving objects
# TODO new room for testing moving objects. Also do move_random and move_to.

# Standard event changes
# TODO standard event for TMD
# TODO better sound effect dang it. For both lab doors and sevahrik

# Changes made:
#  - 'go_to' commands now accept a new argument: prev_coord. If set to True, will cause the object to move to their
#  target's previous  cooridnate rather than its current one
#  - Sprite collection objects can now be given the 'move_to' and 'move_from' properties. This will override the
#  properties from their sprite collection.
#  - 'go_to' resolution now uses code evaluation in determining the target
