import renpy  # noqa


class ControlCommand(object):
    untimed_types = {"change_stand_animation", "change_move_animation", "change_movement_speed", "go_custom"}

    def __init__(self, command_type, **kwargs):
        """
        This class represents a single command, to be given to a sprite to change its position or animation.
        :param str command_type: the type of the command.
        """
        self.type = command_type

        self.properties = {}

        for property_name in kwargs:
            self.properties[property_name] = kwargs[property_name]

    @property
    def turn_and_move(self):
        """
        :return: Used for 'go' commands, specifies whether to turn and move in a single command (if true), or do just
        one of the two. True by default.
        :rtype: bool
        """
        if 'turn_and_move' in self.properties:
            return self.properties['turn_and_move']
        else:
            return True

    @property
    def stand_on_turn(self):
        """
        :return: Used for the 'turn' commands, specifying whether to switch to the a stand animation when turning. True
        by default.
        :rtype: bool
        """
        if 'stand_on_turn' in self.properties:
            return self.properties['stand_on_turn']
        else:
            return True

    @property
    def animate_invalid_move(self):
        """
        :return: Used for the 'move' command, specifying whether to continue using the walk animation when a move is
        deemed invalid. True by default.
        :rtype: bool
        """
        if 'animate_invalid_move' in self.properties:
            return self.properties['animate_invalid_move']
        else:
            return True

    @property
    def pop_invalid_move(self):
        """
        :return: Used for the 'move command', specifying whether to remove a command from the stack when invalid,
        or whether to retry it on the next move.
        :rtype: bool
        """
        if 'pop_invalid_move' in self.properties:
            return self.properties['pop_invalid_move']
        else:
            return False

    @property
    def quantity(self):
        """
        :return: Used in various functions, to specify a numeric input value.
        :rtype: float
        """
        if 'quantity' in self.properties:
            return self.properties['quantity']
        else:
            return None

    @property
    def animation_name(self):
        """
        :return: Used in the play_animation, change_stand_animation and change_move_animation commands, to provide an
        animation name.
        :rtype: str
        """
        if 'animation_name' in self.properties:
            return self.properties['animation_name']
        else:
            return ""

    @property
    def animation_speed(self):
        """
        :return: Used in the play_animation function, to provide an animation speed.
        :rtype: float
        """
        if 'animation_speed' in self.properties:
            return self.properties['animation_speed']
        else:
            return None

    @property
    def set_movement_time(self):
        """
        :return: Whether this animation updates the most recent movement time (if not, causes next command to execute
        immediately). Used by all commands.
        :rtype bool
        """
        if 'set_movement_time' in self.properties:
            return self.properties['set_movement_time']
        else:
            if self.type in self.untimed_types:
                return False
            else:
                return True

    @property
    def target(self):
        """
        :return: Used by the move_to command, determining the target.
        :rtype: Union[tuple, str]
        """
        if 'target' in self.properties:
            return self.properties['target']
        else:
            return 0, 0

    @property
    def prev_coord(self):
        """
        :return: Used by the go_to command, making the movement go to the object's previous target rather than its
        current.
        :rtype: bool
        """
        if 'prev_coord' in self.properties:
            return self.properties['prev_coord']
        else:
            return True

    @property
    def exclude_current_direction(self):
        """
        :return: Used by the 'turn_random' command, whether the current direction should be excluded.
        :rtype: bool
        """
        if 'exclude_current_direction' in self.properties:
            return self.properties['exclude_current_direction']
        else:
            return True

    @property
    def max_x(self):
        """
        :return: Used by the 'go_random' command, determining the bounds of the random movement.
        :rtype: int
        """
        if 'max_x' in self.properties:
            return self.properties['max_x']
        else:
            pink_otm_current_map = renpy.store.__dict__["pink_otm_current_map"]  # Py2
            return pink_otm_current_map.grid_size.x - 1

    @property
    def max_y(self):
        """
        :return: Used by the 'go_random' command, determining the bounds of the random movement.
        :rtype: int
        """
        if 'max_y' in self.properties:
            return self.properties['max_y']
        else:
            pink_otm_current_map = renpy.store.__dict__["pink_otm_current_map"]  # Py2
            return pink_otm_current_map.grid_size.y

    @property
    def min_x(self):
        """
        :return: Used by the 'go_random' command, determining the bounds of the random movement.
        :rtype: int
        """
        if 'min_x' in self.properties:
            return self.properties['min_x']
        else:
            return 0

    @property
    def min_y(self):
        """
        :return: Used by the 'go_random' command, determining the bounds of the random movement.
        :rtype: int
        """
        if 'min_y' in self.properties:
            return self.properties['min_y']
        else:
            return 0

    @property
    def single_step(self):
        """
        :return: Used by the go_to command. If True, causes the movement to only cause a single step.
        :rtype: bool
        """
        if 'single_step' in self.properties:
            return self.properties['single_step']
        else:
            return False

    @property
    def never_repeat(self):
        """
        :return: Prevents the command from being repeated. Used if one command spawns more commands.
        :rtype: bool
        """
        if 'never_repeat' in self.properties:
            return self.properties['never_repeat']
        else:
            return False

    @property
    def move_by_x(self):
        """
        :return: Used by the go_custom command, setting the x movement in the next movement.
        :rtype: int
        """
        if 'move_by_x' in self.properties:
            return self.properties['move_by_x']
        else:
            return 0

    @property
    def move_by_y(self):
        """
        :return: Used by the go_custom command, setting the y movement in the next movement.
        :rtype: int
        """
        if 'move_by_y' in self.properties:
            return self.properties['move_by_y']
        else:
            return 0

    @property
    def movement_time(self):
        """
        :return: Used to set a custom movement_time for only a single movement.
        :rtype: float
        """
        if 'movement_time' in self.properties:
            return self.properties['movement_time']
        else:
            return None

    @property
    def instant(self):
        """
        :return: When this command is consumed, the code immediately starts consuming the next command, without waiting
        for a tick to pass. WARNING: This can cause infinite recursion errors, and nasty crashes. Should only be used
        if you really know what you are doing.
        :rtype: bool
        """
        if 'instant' in self.properties:
            return self.properties['instant']
        else:
            return None
