import renpy  # noqa
from pink_engine.orthogonal_tiled_map_commands import ControlCommand  # noqa


class ParallelProcess:
    def __init__(self):
        pass

    @staticmethod
    def get_trans_ratio(pc_coords, full_area, trans_area):
        # The random float translations are due to python 2 turning int divisions into ints.
        if trans_area[0] <= pc_coords.x <= full_area[0] and (full_area[0] - trans_area[0]) != 0:
            x_ratio = (pc_coords.x - trans_area[0]) / (float(full_area[0]) - trans_area[0])
        elif (trans_area[1] - full_area[1]) != 0:
            x_ratio = (trans_area[1] - pc_coords.x) / (float(trans_area[1]) - full_area[1])
        else:
            x_ratio = 1.0

        if trans_area[2] <= pc_coords.y <= full_area[2] and (full_area[2] - trans_area[2]) != 0:
            y_ratio = (pc_coords.y - trans_area[2]) / (float(full_area[2]) - trans_area[2])
        elif (trans_area[3] - full_area[3]) != 0:
            y_ratio = (trans_area[3] - pc_coords.y) / (float(trans_area[3]) - full_area[3])
        else:
            y_ratio = 1.0

        min_ratio = min(x_ratio, y_ratio)
        return min_ratio

    def on_init(self):
        pass

    def per_tick(self, st):
        pass


class SoundPlayAreas(ParallelProcess):
    def __init__(self, area_music_tuples, default_music, channel="music"):
        ParallelProcess.__init__(self)  # Py2

        # List of length 2 tuples.
        # First element of the tuple is a length 4 tuple, marking the bounds of the area.
        # Second element of the tuple is the music to play in the given area.
        self.area_music_tuples = area_music_tuples
        self.default_music = default_music
        self.channel = channel

    def per_tick(self, st):
        pink_otm_current_pc = renpy.store.pink_otm_current_pc  # Py 2
        pc_coords = pink_otm_current_pc.central_coord

        for area, music_path in self.area_music_tuples:
            if area[0] <= pc_coords.x <= area[1] and area[2] <= pc_coords.y <= area[3]:
                if music_path == "stop":
                    renpy.exports.music.stop(channel=self.channel)
                    return
                elif music_path != "continue":
                    renpy.exports.music.play(music_path, loop=True, if_changed=True, channel=self.channel)
                    return
        renpy.exports.music.play(self.default_music, loop=True, if_changed=True, channel=self.channel)


class SoundPanAreas(ParallelProcess):
    def __init__(self, area_pan_tuples, default_pan, channel="music"):
        ParallelProcess.__init__(self)  # Py2

        # List of length 3 tuples.
        # First element of the tuple is also a length 4 tuple, marking the area where the given pan should exist
        # Second element of the tuple is a length 4, marking the area where the pan should start to transition to
        # the given pan.
        # Third element of the tuple is the given pan.
        self.area_pan_tuples = area_pan_tuples
        self.default_pan = default_pan
        self.channel = channel

    def per_tick(self, st):
        pink_otm_current_pc = renpy.store.pink_otm_current_pc  # Py 2
        pc_coords = pink_otm_current_pc.central_coord

        for full_area, trans_area, pan_level in self.area_pan_tuples:
            if full_area[0] <= pc_coords.x <= full_area[1] and full_area[2] <= pc_coords.y <= full_area[3]:
                renpy.exports.music.set_pan(pan_level, delay=0.1, channel=self.channel)
                return
            elif trans_area[0] <= pc_coords.x <= trans_area[1] and trans_area[2] <= pc_coords.y <= trans_area[3]:
                trans_ratio = self.get_trans_ratio(pc_coords, full_area, trans_area)

                final_pan = self.default_pan + (pan_level - self.default_pan) * trans_ratio
                renpy.exports.music.set_pan(final_pan, delay=0.1, channel=self.channel)
                return

        renpy.exports.music.set_pan(self.default_pan, delay=0.1, channel=self.channel)


class SoundVolumeAreas(ParallelProcess):
    def __init__(self, area_volume_tuples, default_volume, channel="music"):
        ParallelProcess.__init__(self)  # Py2

        # List of length 3 tuples.
        # First element of the tuple is also a length 4 tuple, marking the area where the given volume should exist
        # Second element of the tuple is a length 4, marking the area where the volume should start to transition to
        # the given pan.
        # Third element of the tuple is the given volume.
        self.area_volume_tuples = area_volume_tuples
        self.default_volume = default_volume
        self.channel = channel

    def on_init(self):
        renpy.exports.music.set_volume(self.default_volume, channel=self.channel)

    def per_tick(self, st):
        pink_otm_current_pc = renpy.store.pink_otm_current_pc  # Py 2
        pc_coords = pink_otm_current_pc.central_coord

        for full_area, trans_area, volume_level in self.area_volume_tuples:
            if full_area[0] <= pc_coords.x <= full_area[1] and full_area[2] <= pc_coords.y <= full_area[3]:
                renpy.exports.music.set_volume(volume_level, delay=0.1, channel=self.channel)
                return
            elif trans_area[0] <= pc_coords.x <= trans_area[1] and trans_area[2] <= pc_coords.y <= trans_area[3]:
                trans_ratio = self.get_trans_ratio(pc_coords, full_area, trans_area)

                final_volume = self.default_volume + (volume_level - self.default_volume) * trans_ratio
                renpy.exports.music.set_volume(final_volume, delay=0.1, channel=self.channel)
                return

        renpy.exports.music.set_volume(self.default_volume, delay=0.1, channel=self.channel)


class ZoomAreas(ParallelProcess):
    def __init__(self, area_zoom_tuples, default_zoom, default_zoom_time):
        ParallelProcess.__init__(self)  # Py2

        # List of length 4 tuples.
        # First element of the tuple is also a length 4 tuple, marking the area where the given zoom should exist
        # Second element of the tuple is a length 4 tuple, marking the area where the zoom should start to transition to
        # the given zoom.
        # Third element of the tuple is the given zoom.
        # Fourth element of the tuple is the given zoom time.
        self.area_zoom_tuples = area_zoom_tuples
        self.default_zoom = default_zoom
        self.default_zoom_time = default_zoom_time

    def per_tick(self, st):
        pink_otm_current_pc = renpy.store.pink_otm_current_pc  # Py 2
        pink_otm_current_camera = renpy.store.pink_otm_current_camera  # Py 2
        pc_coords = pink_otm_current_pc.central_coord

        for full_area, trans_area, zoom_level, zoom_time in self.area_zoom_tuples:
            zoom_time = min(pink_otm_current_pc.movement_speed - 0.03, zoom_time)

            if full_area[0] <= pc_coords.x <= full_area[1] and full_area[2] <= pc_coords.y <= full_area[3]:
                pink_otm_current_camera.smooth_zoom(zoom_level, zoom_time)
                return
            elif trans_area[0] <= pc_coords.x <= trans_area[1] and trans_area[2] <= pc_coords.y <= trans_area[3]:
                trans_ratio = self.get_trans_ratio(pc_coords, full_area, trans_area)

                final_zoom = self.default_zoom + (zoom_level - self.default_zoom) * trans_ratio
                pink_otm_current_camera.smooth_zoom(final_zoom, zoom_time)
                return

        default_zoom_time = min(pink_otm_current_pc.movement_speed - 0.03, self.default_zoom_time)
        pink_otm_current_camera.smooth_zoom(self.default_zoom, default_zoom_time)


class VarAreas(ParallelProcess):
    def __init__(self, area_var_tuples):
        ParallelProcess.__init__(self)  # Py2

        # List of length 3 tuples.
        # First element of the tuple is a length 4 tuple, marking the boundaries of the area.
        # Second element of the tuple is the name of the variable. It should be a variable in the renpy global store
        # Third element of the tuple is the value the variable should be set to on entering the area.
        self.area_var_tuples = area_var_tuples

    def per_tick(self, st):
        pink_otm_current_pc = renpy.store.pink_otm_current_pc  # Py 2
        pc_coords = pink_otm_current_pc.central_coord

        for var_area, var_name, var_value in self.area_var_tuples:
            if var_area[0] <= pc_coords.x <= var_area[1] and var_area[2] <= pc_coords.y <= var_area[3]:
                setattr(renpy.store, var_name, var_value)


class CameraSwitchAreas(ParallelProcess):
    def __init__(self, area_camera_tuples, default_camera=None, default_zoom=1.0):
        ParallelProcess.__init__(self)  # Py2

        # List of length 3 tuples.
        # First element of the tuple is a length 4 tuple, marking the boundaries of the area.
        # Second element is the ref_name of the object the camera should switch to
        # Third element is the zoom that should be used for this camera.
        self.area_camera_tuples = area_camera_tuples
        self.default_camera = default_camera
        self.default_zoom = default_zoom

    def per_tick(self, st):
        pink_otm_current_pc = renpy.store.pink_otm_current_pc  # Py 2
        pink_otm_current_camera = renpy.store.pink_otm_current_camera  # Py 2
        pc_coords = pink_otm_current_pc.central_coord
        zoom_time = pink_otm_current_pc.movement_speed - 0.04

        for camera_area, ref_camera, zoom in self.area_camera_tuples:
            if camera_area[0] <= pc_coords.x <= camera_area[1] and camera_area[2] <= pc_coords.y <= camera_area[3]:
                pink_otm_current_camera.switch_target(getattr(renpy.store, ref_camera))
                pink_otm_current_camera.smooth_zoom(zoom, zoom_time)
                return

        if self.default_camera is None:
            pink_otm_current_camera.recenter_on_player()
        else:
            pink_otm_current_camera.switch_target(getattr(renpy.store, self.default_camera))
        pink_otm_current_camera.smooth_zoom(self.default_zoom, zoom_time)


class AnimationAreas(ParallelProcess):
    def __init__(self, area_animation_tuples):
        ParallelProcess.__init__(self)  # Py2

        # List of length 3 tuples
        # First element of the tuple is a length 4 tuple, marking the boundaries of the area.
        # Second element is a list of length 3 tuples, which indicate the animation to play/set on entering the
        # area
        # Third element is a list of length 3 tuples, which indicate the animation to play/set on leaving the area.
        # For the lists of length 3 tuples, the first element is a string with the value 'play', 'set stand', or
        # 'set move', which indicates what kind of action you want. The second element is the name of the animation.
        # The third element is the refname of the object that should have the animation set/played.
        self.area_animation_tuples = area_animation_tuples

        self.in_areas = set()

    def per_tick(self, st):
        pink_otm_current_pc = renpy.store.pink_otm_current_pc  # Py 2
        pc_coords = pink_otm_current_pc.central_coord

        for animation_area, enter_list, leave_list in self.area_animation_tuples:
            if (
                    animation_area not in self.in_areas and
                    animation_area[0] <= pc_coords.x <= animation_area[1] and
                    animation_area[2] <= pc_coords.y <= animation_area[3]
            ):
                # Entering area
                self.in_areas.add(animation_area)
                self._run_animation_list(enter_list)

            elif (
                    animation_area in self.in_areas and
                    not (animation_area[0] <= pc_coords.x <= animation_area[1] and
                         animation_area[2] <= pc_coords.y <= animation_area[3])
            ):
                # Leaving area
                self.in_areas.remove(animation_area)
                self._run_animation_list(leave_list)

    @staticmethod
    def _run_animation_list(animation_list):
        for refname, command in animation_list:
            target = getattr(renpy.store, refname)  # Py 2

            target.control_stack.append(command)


class VarTimes(ParallelProcess):
    def __init__(self, var_time_tuples):
        ParallelProcess.__init__(self)  # Py2

        # List of length 4 tuples. Tuple 1 is time in seconds, tuple 2 is name of the variable, tuple 3 is a string
        # that represents the action to be taken (set, add, sub), tuple 4 is the value to set, add or sub.
        self.var_time_tuples = var_time_tuples
        self.last_trigger_st = 0

    def per_tick(self, st):
        if len(self.var_time_tuples) > 0 and st >= self.last_trigger_st + self.var_time_tuples[0][0]:
            time, var_name, action, var_value = self.var_time_tuples.pop(0)
            if action == "set":
                setattr(renpy.store, var_name, var_value)
            elif action == "add":
                setattr(renpy.store, var_name, getattr(renpy.store, var_name) + var_value)
            elif action == "sub":
                setattr(renpy.store, var_name, getattr(renpy.store, var_name) - var_value)

            self.last_trigger_st = self.last_trigger_st + time
            self.var_time_tuples.append((time, var_name, action, var_value))


class SmoothZoomController(ParallelProcess):
    def __init__(self):
        ParallelProcess.__init__(self)  # Py2

        # Smooth zoom-related variables
        # self.start_zooming = True
        self.original_xzoom = renpy.store.pink_otm_camera_default_zoom  # type: float
        self.target_xzoom = renpy.store.pink_otm_camera_default_zoom  # type: float
        self.original_yzoom = renpy.store.pink_otm_camera_default_zoom  # type: float
        self.target_yzoom = renpy.store.pink_otm_camera_default_zoom  # type: float
        self.start_zoom_st = 0.0
        self.last_st = 0.0  # used by smooth zoom function
        self.zoom_speed = 1.0  # Time in seconds to complete the zoom

    def per_tick(self, st):
        """
        """
        pink_otm_current_camera = renpy.store.pink_otm_current_camera
        self.last_st = st

        zoomed = False
        if pink_otm_current_camera.xzoom != self.target_xzoom:
            if self.start_zoom_st > st:
                self.start_zoom_st = 0.00001
            st_diff = st - self.start_zoom_st
            st_fraction = min(1.0, st_diff / self.zoom_speed)
            new_xzoom = self.original_xzoom - (self.original_xzoom - self.target_xzoom) * st_fraction
            pink_otm_current_camera.xzoom = new_xzoom
            zoomed = True

        if pink_otm_current_camera.yzoom != self.target_yzoom:
            if self.start_zoom_st > st:
                self.start_zoom_st = 0.00001
            st_diff = st - self.start_zoom_st
            st_fraction = min(1.0, st_diff / self.zoom_speed)
            new_yzoom = self.original_yzoom - (self.original_yzoom - self.target_yzoom) * st_fraction
            pink_otm_current_camera.yzoom = new_yzoom
            zoomed = True

        if zoomed:
            renpy.exports.restart_interaction()
