import renpy  # noqa
import pygame  # noqa
import json
from pink_engine.commons import Coord
from pink_engine.tileset import Tileset, TilesetTile, TilesetAnimatedTile, pink_tileset_dict
from math import ceil, floor
from collections import OrderedDict
import os

# ======================================================================================================================
# =============================================GLOBAL SETTINGS==========================================================
# ======================================================================================================================

LAYER_PRIORITY = {
    'sub': 0,
    'dynamic': 1,
    'super': 2,
    'move': 3,
    'interaction': 3}  # TODO split render and nonrender layers


# ======================================================================================================================
# =============================================VISUAL===================================================================
# ======================================================================================================================


class TiledMap(renpy.exports.Displayable):
    def __init__(self, game_object, **kwargs):
        """
        The displayable object that represents a single pink engine map.
        :param TiledMapGame game_object: The game that this object displays.
        """
        renpy.exports.Displayable.__init__(self, **kwargs)  # Py2

        # The associated gameplay object. Should set a number of variables.
        self.game_object = game_object

        # The grid of all displayable objects.
        self.render_grid = []
        self.render_grid_x_size = ceil(self.game_object.image_size.x / self.game_object.tile_size.x)
        self.render_grid_y_size = ceil(self.game_object.image_size.y / self.game_object.tile_size.y)
        for x in range(self.render_grid_x_size):
            self.render_grid.append([])
            for y in range(self.render_grid_y_size):
                self.render_grid[x].append({})

        # All elements on the map have their own unique id. This allows us to make sure each renders only once,
        # as well as letting us find and delete them with relative ease.
        self.next_id = 0
        self.map_elements = {}

        #

    def get_next_id(self):
        """
        :return: The id that should be used for the next displayable element
        :rtype: int
        """
        next_id = self.next_id
        self.next_id += 1
        return next_id

    def remove_from_grid(self, element):
        """
        Removes the given element from this object's render grid
        :param TiledMapElement element: The element to remove from the render grid
        """
        for column in element.get_columns():
            for row in element.get_rows():
                if 0 <= row < self.render_grid_y_size and 0 <= column < self.render_grid_x_size:
                    self.render_grid[column][row].pop(element.element_id)

    def add_to_grid(self, element):
        """
        Adds the given element to this object's render grid
        :param TiledMapElement element: The element to add to the render grid
        """
        for column in element.get_columns():
            for row in element.get_rows():
                if 0 <= row < self.render_grid_y_size and 0 <= column < self.render_grid_x_size:
                    self.render_grid[column][row][element.element_id] = element

    def add_element(self, x, y, layer, game_object):
        """
        Adds the given game object at the given x and y on the given layer
        :param int x: The x (in pixels) at which the given object's image is displayed
        :param int y: The y (in pixels) at which the given object's image is displayed
        :param TiledMapGameLayer layer: The layer the object exists on
        :param TiledMapGameTile|TiledMapGameObject game_object: The given game object
        :return: The displayable ID of the added object
        :rtype: int
        """
        element_id = self.get_next_id()
        new_element = TiledMapElement(self, element_id, x, y, layer, game_object)
        self.map_elements[element_id] = new_element
        self.add_to_grid(new_element)

        if game_object.single_display:
            game_object.displayable_id = element_id

        return element_id

    def remove_element(self, element_id):
        """
        Removes the element with the given ID from this displayable
        :param int element_id: The given element id
        """
        element = self.map_elements[element_id]
        self.remove_from_grid(element)
        self.map_elements.pop(element_id)

        if element.game_object.single_display:
            element.game_object.displayable_id = None

    def move_element(self, element_id, new_x, new_y):
        """
        Moves the element with the given ID to the given new x and y
        :param int element_id: The given element id
        :param int new_x: The new x (in pixels)
        :param int new_y: The new y (in pixels)
        """
        element = self.map_elements[element_id]
        self.remove_from_grid(element)
        element.x = new_x
        element.y = new_y
        element.set_priority_tuple()
        self.add_to_grid(element)

    def render(self, width, height, st, at):
        """
        :param int width:
        :param int height:
        :param float st: The shown time (in seconds).
        :param float at: The animation time (in seconds).
        :rtype: renpy.exports.Render
        """
        renpy.exports.redraw(self, 0)

        # initiate render object
        render = renpy.exports.Render(width, height)

        if renpy.store.pink_otm_current_map is None:  # prevents crash on completion of game
            return render

        self.game_object.camera.center_on_target()
        objects_in_frame = self.get_objects_in_frame()
        objects_in_frame = self.sort_objects(objects_in_frame)

        for child in objects_in_frame:
            render.place(child, child.x, child.y)

        t = renpy.display.transform.Transform()
        renpy.exports.render(t, width, height, st, at)  # child render

        return render

    def get_objects_in_frame(self):
        """
        :return: a list of objects currently in frame.
        :rtype: list
        """
        camera = self.game_object.camera
        xzoom, yzoom = camera.xzoom, camera.yzoom

        xzoom_mod = (1 / xzoom)
        yzoom_mod = (1 / yzoom)
        min_x_bound = int((camera.x.value * xzoom_mod) / self.game_object.tile_size.x)
        max_x_bound = int(
            ((camera.x.value + self.game_object.camera.viewport_width) * xzoom_mod) / self.game_object.tile_size.x)
        min_y_bound = int((camera.y.value * yzoom_mod) / self.game_object.tile_size.y)
        max_y_bound = int(
            ((camera.y.value + self.game_object.camera.viewport_height) * yzoom_mod) / self.game_object.tile_size.y)

        # Slice render_grid to get matrix of objects currently in frame.
        slices = []
        frame_matrix = self.render_grid[min_x_bound:max_x_bound + 1]
        for sliced_row in frame_matrix:
            slices.append(sliced_row[min_y_bound:max_y_bound + 1])

        # Translates matrix to dictionary, using object id to get rid of doubles.
        objects_in_frame = {}
        for sliced_row in slices:
            for sliced_tile in sliced_row:
                for object_id, map_object in sliced_tile.items():
                    objects_in_frame[object_id] = map_object

        return list(objects_in_frame.values())

    @staticmethod
    def sort_objects(objects_to_sort):
        """
        :param list objects_to_sort: The list of objects to sort
        :return: The given list of objects sorted according to their priority dictionary, so that they are in the order
        in which they should be rendered.
        :rtype: list
        """
        priority_dict = {}
        for map_object in objects_to_sort:
            priority_tuple = map_object.priority_tuple
            priority_dict[priority_tuple] = map_object

        sorted_objects = []
        for key in sorted(priority_dict):
            sorted_objects.append(priority_dict[key])

        return sorted_objects

    def visit(self):
        return list(self.map_elements.values())


class TiledMapElement(renpy.exports.Displayable):
    def __init__(self, parent_map, element_id, x, y, layer, game_object, **kwargs):
        """
        A single displayable object in the TiledMap object.
        :param TiledMapGame parent_map: The game map which this element is part of the display of
        :param element_id:
        :param x: The distance in pixels between this object's image's left side and the left border of the map
        :param y: The distance in pixels between this object's image's left side and the left border of the map
        :param layer: The layer on which this object exists
        :param game_object: The game object which this element is displaying
        """
        renpy.exports.Displayable.__init__(self, **kwargs)  # Py2

        self.parent_map = parent_map
        self.element_id = element_id
        self.x = x
        self.y = y
        self.game_object = game_object
        self.layer = layer

        if self.game_object.parent.orientation == "orthogonal" and self.layer.layer_type == 'dynamic':
            self._priority_tuple = (
                LAYER_PRIORITY[self.layer.layer_type], self.render_y, self.layer.z_order, self.x, self.element_id)
        else:
            self._priority_tuple = (
                LAYER_PRIORITY[self.layer.layer_type], self.layer.z_order, self.render_y, self.x, self.element_id)


    def render(self, width, height, st, at):  # noqa
        """
        :param int width:
        :param int height:
        :param float st: The shown time (in seconds).
        :param float at: The animation time (in seconds).
        :rtype: renpy.exports.Render
        """
        child = self.game_object.image

        if self.game_object.render_dict == {}:
            child_render = renpy.exports.load_image(child)
        else:
            t = renpy.display.transform.Transform(child=child, **self.game_object.render_dict)
            child_render = renpy.exports.render(t, width, height, st, at)

        if self.game_object.is_animated:
            renpy.exports.redraw(self, 0.05)
            # TODO issues when next frame time is changed. Maybe create class for keeping track?
            # renpy.exports.redraw(self, st - self.game_object.next_frame_time)
        return child_render

    def get_columns(self):
        """
        :return: The range of columns of the render grid occupied by this element
        :rtype: range
        """
        start_x_coord = int(self.x / self.parent_map.game_object.tile_size.x)
        end_x_coord = int((self.x + self.game_object.width - 1) / self.parent_map.game_object.tile_size.x)
        return range(start_x_coord, end_x_coord + 1)

    def get_rows(self):
        """
        :return: The range of rows of the render grid occupied by this element
        :rtype: range
        """
        start_y_coord = int(self.y / self.parent_map.game_object.tile_size.y)
        end_y_coord = int((self.y + self.game_object.height - 1) / self.parent_map.game_object.tile_size.y)
        return range(start_y_coord, end_y_coord + 1)

    @property
    def priority_tuple(self):
        """
        :return: This object's priority tuple (a tuple determining the order in which objects are rendered)
        :rtype: tuple
        """
        return self._priority_tuple

    @property
    def render_y(self):
        """
        :return: The y coordinate at which this object is regarded as being placed for the sake of determining its
        render order on dynamic layers
        :rtype: int
        """
        try:
            return self.game_object.occultates_as_y
        except AttributeError:
            pass
        if self.game_object.parent.orientation == "orthogonal":
            return floor((self.y + self.game_object.height - 2) / self.game_object.parent.tile_size.y)
        else:
            return self.y + self.game_object.height - 1

    def set_priority_tuple(self):
        """
        Updates this object's priority tuple
        """
        if self.game_object.parent.orientation == "orthogonal" and self.layer.layer_type == 'dynamic':
            self._priority_tuple = (
                LAYER_PRIORITY[self.layer.layer_type], self.render_y, self.layer.z_order, self.x, self.element_id)
        else:
            self._priority_tuple = (
                LAYER_PRIORITY[self.layer.layer_type], self.layer.z_order, self.render_y, self.x, self.element_id)

# TODO OrthogonalTiledMapPolygon


# ======================================================================================================================
# ===========================================GAMEPLAY===================================================================
# ======================================================================================================================


# Base Images
class TiledMapGameTile(object):
    is_conditional = False
    is_animated = False
    single_display = False  # If True, Indicates that this object corresponds to only a single visual object.

    def __init__(self, parent_map, parent_tile):
        """
        A single tile on a Tiled Map (note, all instances of a tile in a game refer to the same object, whereas all
        objects on a game map have their own object)
        :param TiledMapGame parent_map: The game map that uses this tile
        :param TilesetTile parent_tile: The tile in the tileset that this tile is using the properties of.
        """
        self.parent = parent_map
        self.image = filepath_to_reg_image(parent_tile.image)
        self.parent.registered_images.add(self.image)

        self.height = parent_tile.image_height
        self.width = parent_tile.image_width
        self.properties = parent_tile.properties.copy()

    @property
    def render_dict(self):
        """
        :return: The render dictionary that should be used to render displayable elements linked to this tile
        :rtype: dict
        """
        return {}


class TiledMapGameTileAnimated(TiledMapGameTile):
    is_animated = True

    def __init__(self, parent_map, parent_tile):
        """
        A single tile on a Tiled Map (note, all instances of a tile in a game refer to the same object, whereas all
        objects on a game map have their own object)
        :param TiledMapGame parent_map: The game map that uses this tile
        :param TilesetAnimatedTile parent_tile: The tile in the tileset that this tile is using the properties of.
        """
        TiledMapGameTile.__init__(self, parent_map, parent_tile)

        self.frames = OrderedDict()
        for animation_time, frame_id in parent_tile.frames.items():
            animation_time = float(animation_time) / 1000
            frame_image = filepath_to_reg_image(parent_tile.parent.get_tile(frame_id).image)
            self.frames[animation_time] = frame_image
            self.parent.registered_images.add(frame_image)

        self.next_frame_key = next(iter(self.frames))
        self.next_frame_time = next(iter(self.frames))
        self.image = self.frames[self.next_frame_key]
        self.latest_frame_gt = self.parent.last_gt
        self.paused = False

    def pause_animation(self):
        """
        Pauses the animation on a map freeze.
        """
        self.paused = True

    def unpause_animation(self):
        """
        Unpauses the animation on a map unfreeze
        """
        self.paused = False

    def increment_gt(self, gt_diff):
        """
        Delays the time at which the next animated frame appears by the given quantity.
        :param float gt_diff: the quantity of time by which to delay the next frame (in seconds)
        """
        self.next_frame_time += gt_diff

    def update_frame(self, gt):
        """
        Updates this object's displayed frame to the frame it should have at the given game time.
        :param float gt: Game time in seconds.
        """
        self.latest_frame_gt = gt
        if not self.paused:
            while gt > self.next_frame_time:
                prev_frame_key = self.next_frame_key

                old_frame = self.frames.pop(self.next_frame_key)
                self.frames[self.next_frame_key] = old_frame

                self.next_frame_key = next(iter(self.frames))
                if self.next_frame_key > prev_frame_key:
                    self.next_frame_time = self.next_frame_time + self.next_frame_key - prev_frame_key
                else:
                    self.next_frame_time = self.next_frame_time + self.next_frame_key

            self.image = self.frames[self.next_frame_key]

    def st_reset_frames(self, st):
        """
        Rebases this animation's next frame time based on the given st. This function was designed to
        be used after st resets when displaying a map.
        :param float st: Game time in seconds.
        """
        self.next_frame_time = self.next_frame_time - self.latest_frame_gt + st


class TiledMapSpriteCollectionElement(object):
    def __init__(self, element_dict, parent_sprite_collection):
        """
        This class represents a single element in a SpriteCollection. It can be either a still image, or
        an animation, depending on the number of images in the element.
        :param dict element_dict: A json-derived dictionary that describes the contents of this element.
        :param TiledMapSpriteCollection parent_sprite_collection: The sprite collection this element is a part of.
        """
        self.parent_sprite_collection = parent_sprite_collection

        self.name = element_dict["name"]
        self.images = []
        for image_path in element_dict["images"]:
            reg_path = filepath_to_reg_image(image_path)
            self.images.append(reg_path)
        self.default_speed = element_dict["default_speed"]  # type: float
        self.image_width = element_dict["image_width"]  # type: int
        self.image_height = element_dict["image_height"]  # type: int

        self.properties = {}
        for element_property in element_dict['properties']:
            self.properties[element_property['name']] = element_property['value']


class TiledMapSpriteCollection(object):  # TODO Mutual parent class for this and tile
    is_animated = True

    def __init__(self, json_file, start_time=0.0):
        """
        This class represents all images that belong to a single on-screen sprite, either a character or an object.
        These images are collected into elements, forming either still images or animations, depending on the number of
        images in the element. One element is the "current_animation", which is the element that is currently being
        displayed by the sprite.
        :param json_file: The file that contains this sprite collection's data.
        :param float start_time: the start time for this sprite collection.
        """
        sprite_collection_dict = json.load(json_file)

        # Initialize individual animations
        self.data = {}
        for element_dict in sprite_collection_dict['animations']:
            self.data[element_dict["name"]] = TiledMapSpriteCollectionElement(
                element_dict=element_dict, parent_sprite_collection=self)

        # Animation variables
        self.current_animation = None
        self.frames = OrderedDict()
        self.next_frame_time = start_time
        self.next_frame_key = 0.0
        self.current_speed = 1.0
        self._image = None

        # Pause variables
        self.paused = False
        self.latest_frame_gt = start_time

    def get_all_images(self):
        """
        :return: a list of all image names displayed by this sprite collection.
        :rtype: list
        """
        all_images = []
        for animation in self.data.values():
            for frame_image in animation.images:
                all_images.append(frame_image)
        return all_images

    def increment_gt(self, gt_diff):
        """
        Delays the time at which the next animated frame appears by the given quantity.
        :param float gt_diff: the quantity of time by which to delay the next frame (in seconds)
        """
        self.next_frame_time += gt_diff

    def pause_animation(self):
        """
        Pauses the animation on a map freeze.
        """
        self.paused = True

    def unpause_animation(self):
        """
        Unpauses the animation on a map unfreeze.
        """
        self.paused = False

    def update_frame(self, gt):
        """
        Updates this object's displayed frame to the frame it should have at the given game time.
        :param float gt: Game time in seconds.
        """
        self.latest_frame_gt = gt
        if not self.paused and self.current_animation is not None:
            while gt > self.next_frame_time:
                prev_frame_key = self.next_frame_key

                old_frame = self.frames.pop(prev_frame_key)
                self.frames[prev_frame_key] = old_frame

                self.next_frame_key = next(iter(self.frames))
                if self.next_frame_key > prev_frame_key:
                    self.next_frame_time = self.next_frame_time + self.next_frame_key - prev_frame_key
                else:
                    self.next_frame_time = self.next_frame_time + self.next_frame_key

            self._image = self.frames[self.next_frame_key]

    def st_reset_frames(self, st):
        """
        Rebases this animation's next frame time based on the given st. This function was designed to
        be used after st resets when displaying a map.
        :param float st: Game time in seconds.
        """
        self.next_frame_time = self.next_frame_time - self.latest_frame_gt + st

    @staticmethod
    def _get_frames(animation, time_per_frame):
        """
        Returns an ordered dictionary that represents the st key/frame value combinations to play the given animation
        at the given speed from the given start time.
        :param TiledMapSpriteCollectionElement animation: The animation in question
        :param float time_per_frame: the time in seconds between frames of the animation
        :rtype: OrderedDict
        """
        frame_time = time_per_frame
        frames = OrderedDict()
        for image_path in animation.images:
            frames[frame_time] = image_path
            frame_time += time_per_frame
        return frames

    def set_animation(self, element_name, speed=None):
        """
        Changes the current_animation to the element with the given name. Can also override default speed.
        :param str element_name: The name of the element to change the current_animation to.
        :param float speed: Sets the speed for the animation. If None, the animation will play at its default speed.
        None by default.
        """
        # Do nothing if already displaying this animation
        if self.current_animation and speed == self.current_speed and element_name == self.current_animation.name:
            return

        if element_name in self.data:
            self.current_animation = self.data[element_name]
            if speed is None:
                self.current_speed = self.current_animation.default_speed
            else:
                self.current_speed = speed
            self.frames = self._get_frames(self.current_animation, self.current_speed)

            self.next_frame_key = next(iter(self.frames))
            self.next_frame_time = self.latest_frame_gt + next(iter(self.frames))
            self._image = self.frames[self.next_frame_key]

        else:
            self.current_animation = None
            self.frames = OrderedDict()
            self.next_frame_time = 0.0
            self.next_frame_key = 0.0
            self.current_speed = 1.0
            self._image = None

    @property
    def width(self):
        """
        :return: Width of the image, Inherited from current animation of set, 0 if not.
        :rtype: int
        """
        if self.current_animation:
            return self.current_animation.image_width
        else:
            return 0

    @property
    def height(self):
        """
        :return: Height of the image, Inherited from current animation of set, 0 if not.
        :rtype: int
        """
        if self.current_animation:
            return self.current_animation.image_height
        else:
            return 0

    @property
    def image(self):
        """
        Retrieves the ren'py name for the image currently being portrayed by this collection.
        :return: The path to the image currently being portrayed by this collection.
        :rtype: str|None
        """
        if self.current_animation:
            return self._image
        else:
            return None


# Objects
class TiledMapGameObject(object):
    single_display = True

    def __init__(self, parent_rpg_tile, x, y, width, height, layer, properties):
        """
        Represents a single object on the map. Unlike tiles, objects can have individual settings, allowing them to
        move, be scaled, have conditional properties, etc.
        :param TiledMapGameTile|TiledMapSpriteCollection parent_rpg_tile: The tile or sprite collection that this
        object is displaying.
        :param int x: The distance in pixels between this object's image's left side and the left border of the map
        :param int y: The distance in pixels between this object's image's left side and the left border of the map
        :param int width: The width in pixels of this object
        :param int height: The height in pixels of this object
        :param TiledMapGameLayer layer: The layer on which this object exists
        :param dict properties: The dictionary of properties for this object
        """
        self.parent_rpg_tile = parent_rpg_tile
        self.layer = layer
        self.parent = layer.parent

        if self.should_load_consistent(properties):
            old_incarnation = getattr(renpy.store, properties['ref_name'], None)  # type: TiledMapGameObject
            self.x = old_incarnation.x
            self.y = old_incarnation.y
            self._width = old_incarnation.width
            self._height = old_incarnation.height
            self.properties = old_incarnation.properties

        else:
            self.x = x
            self.y = y
            self._width = width
            self._height = height

            self.properties = properties

        self.xzoom, self.yzoom = self._get_zoom()

        if 'conditional' in self.properties:
            self.is_conditional = True
        else:
            self.is_conditional = False

        if parent_rpg_tile.is_animated:
            self.is_animated = True
        else:
            self.is_animated = False

        self.displayable_id = None

        if type(self) is TiledMapGameObject and self.ref_name is not None:
            # Type comparison ensures this only happens at the end of the final loading
            setattr(renpy.store, self.ref_name, self)

    @staticmethod
    def should_load_consistent(properties):
        """
        :param dict properties: the given properties
        :return: Whether an object with the given properties should result in an old instance of said object
        being loaded
        :rtype: bool
        """
        return 'ref_name' in properties and getattr(renpy.store, properties['ref_name'], None) is not None

    def _get_zoom(self):
        """
        :return: A tuple containing the x and y scaling of this object relative to the original tile.
        :rtype: tuple
        """
        xzoom = float(self.width) / float(self.parent_rpg_tile.width)
        yzoom = float(self.height) / float(self.parent_rpg_tile.height)

        return xzoom, yzoom

    @property
    def ref_name(self):
        """
        :return: the reference name used for this object. If set, creates a ren'py variable that corresponds to this
        object, allowing it to be targeted during events.
        :rtype: str|None
        """
        if 'ref_name' in self.properties:
            return self.properties['ref_name']
        else:
            return None

    @property
    def width(self):
        """
        :return: the width of this object in pixels.
        :rtype: int
        """
        return self._width

    @property
    def height(self):
        """
        :return: the height of this object in pixels.
        :rtype: int
        """
        return self._height

    @property
    def render_dict(self):
        """
        :return: The render dictionary that should be used to render displayable elements linked to this object
        :rtype: dict
        """
        if self.xzoom != 1.0 or self.yzoom != 1.0:
            return {'xzoom': self.xzoom, 'yzoom': self.yzoom, 'subpixel': True}
        return {}

    @property
    def image(self):
        """
        Retrieves the ren'py name for the image currently being portrayed by this object.
        :return: The path to the image currently being portrayed by this object.
        :rtype: str|None
        """
        return self.parent_rpg_tile.image

    @property
    def occultates_as_y(self):
        """
        :return: The y coordinate at which this object is regarded as being placed for the sake of determining its
        render order on dynamic layers
        :rtype: int
        """
        if 'occultates_as_y' in self.properties:
            return self.properties['occultates_as_y']
        if self.parent.orientation == "orthogonal":
            return ceil((self.y + self.height - 2) / self.parent.tile_size.y)
        else:
            return self.y + self.height - 1

    @property
    def next_frame_time(self):
        """
        :return: The frame time at which this object should display its next frame.
        """
        return self.parent_rpg_tile.next_frame_time

    def condition_met(self):
        """
        :return: Whether or not this object's conditional properties are currently met
        :rtype: bool
        """
        if not self.is_conditional:
            return True
        return eval(self.properties['conditional'])


class TiledMapGameObjectSpriteCollection(TiledMapGameObject):
    def __init__(
            self, x, y, layer, properties, collection_path
    ):
        """
        :param int x: The distance in pixels between this object's image's left side and the left border of the map
        :param int y: The distance in pixels between this object's image's left side and the left border of the map
        :param TiledMapGameLayer layer: The layer on which this object exists
        :param dict properties: The dictionary of properties for this object
        :param str collection_path: The path to this object's sprite collection
        """
        if self.should_load_consistent(properties):
            old_incarnation = getattr(renpy.store, properties['ref_name'], None)  # type: TiledMapGameObjectSpriteCollection  # noqa
            self.sprite_collection_path = old_incarnation.sprite_collection_path
            self.sprite_collection = old_incarnation.sprite_collection
        else:
            if not hasattr(self, 'sprite_collection'):
                self.sprite_collection_path = collection_path
                self.sprite_collection = TiledMapSpriteCollection(renpy.exports.file(collection_path))

        TiledMapGameObject.__init__(
            self, self.sprite_collection, x, y, self.sprite_collection.width,
            self.sprite_collection.height, layer, properties)

        self.current_animation_name = None  # Set by _set_animation
        if self.should_load_consistent(properties):
            old_incarnation = getattr(renpy.store, properties['ref_name'], None)  # type: TiledMapGameObjectSpriteCollection  # noqa

            self._orientation = old_incarnation._orientation
            self._set_animation(old_incarnation.current_animation_name)
            self.st_reset_frames(0.0)
        else:
            # These need properties to initialize, so are after the parent class initialization
            self._orientation = self.start_orientation
            self._set_animation(self.default_animation)

        self.parent.animated_tiles.append(self.sprite_collection)

        if type(self) is TiledMapGameObjectSpriteCollection and self.ref_name is not None:
            # Type comparison ensures this only happens at the end of the final loading
            setattr(renpy.store, self.ref_name, self)

    def _set_animation(self, animation_name, speed=None):
        """
        This function is called to set the object's current animation set. It calls an oriented animation
        (ex. walk_left) if one is available.
        :param str animation_name: Name of the new animation
        :param float speed: The speed at which to play the animation. None if the default speed should be used.
        """
        if animation_name is None:
            return

        oriented_name = animation_name + "_" + self._orientation
        if oriented_name in self.sprite_collection.data:
            if oriented_name != self.current_animation_name:
                self.current_animation_name = oriented_name
                self.sprite_collection.set_animation(oriented_name, speed=speed)
        else:
            if animation_name != self.current_animation_name:
                self.current_animation_name = animation_name
                self.sprite_collection.set_animation(animation_name, speed=speed)

    def st_reset_frames(self, st):
        """
        :param float st: Game time in seconds.
        :return: Rebases this animation's next frame time based on the given st. This function was designed to
        be used after st resets when displaying a map.
        """
        self.sprite_collection.st_reset_frames(st)

    def _get_zoom(self):
        """
        :return: A tuple containing the x and y scaling of this object relative to the original tile.
        :rtype: tuple
        """
        return 1.0, 1.0

    @property
    def width(self):
        """
        :return: the width of this object in pixels.
        :rtype: int
        """
        return self.sprite_collection.width

    @property
    def height(self):
        """
        :return: the height of this object in pixels.
        :rtype: int
        """
        return self.sprite_collection.height

    @property
    def start_orientation(self):
        """
        :return: the orientation this sprite collection object should start out with.
        :rtype: str
        """
        if 'start_orientation' in self.properties:
            return self.properties['start_orientation']
        else:
            return 'down'

    @property
    def stand_animation(self):
        """
        :return: the name of the animation to display when this object is standing still.
        :rtype: str
        """
        if 'stand_animation' in self.properties:
            return self.properties['stand_animation']
        else:
            return 'stand'

    @stand_animation.setter
    def stand_animation(self, value):
        """
        :param str value: The new stand_animation name
        """
        if (
                self.current_animation_name == self.stand_animation or
                self.current_animation_name == self.stand_animation + "_" + self._orientation
        ):
            currently_standing = True
        else:
            currently_standing = False

        self.properties['stand_animation'] = value

        if currently_standing:
            self._set_animation(value)

    @property
    def move_animation(self):
        """
        :return: the name of the animation to display when this object is moving.
        :rtype: str
        """
        if 'move_animation' in self.properties:
            return self.properties['move_animation']
        else:
            return 'walk'

    @move_animation.setter
    def move_animation(self, value):
        """
        :param str value: The new move_animation name
        """
        if (
                self.current_animation_name == self.move_animation or
                self.current_animation_name == self.move_animation + "_" + self._orientation
        ):
            currently_moving = True
        else:
            currently_moving = False

        self.properties['move_animation'] = value

        if currently_moving:
            self._set_animation(value)

    @property
    def default_animation(self):
        """
        :return: the name of the animation to display when this object is initialized
        :rtype: str
        """
        if 'default_animation' in self.properties:
            return self.properties['default_animation']
        else:
            return self.stand_animation

    @property
    def orientation(self):
        """
        :return: The object's current orientation
        :rtype: str
        """
        return self._orientation

    @orientation.setter
    def orientation(self, value):
        """
        :param str value: Changes the object's orientation to the given orientation
        """
        uvalue = "_" + value
        current_animation_name = self.current_animation_name.\
            replace('_up', uvalue).replace('_down', uvalue).replace('_left', uvalue).replace('_right', uvalue)

        self._orientation = value
        self._set_animation(current_animation_name)

    @property
    def image(self):
        """
        Retrieves the ren'py name for the image currently being portrayed by this collection object.
        :return: The path to the image currently being portrayed by this collection object.
        :rtype: str|None
        """
        if self.sprite_collection is None:
            return None
        else:
            return self.sprite_collection.image

    def switch_sprite_collection(self, target_sprite_collection):
        """
        Switches to a new sprite collection, ensuring that coordinates and animation name are kept the same.
        :param str target_sprite_collection: Path of the new sprite collection.
        :return:
        """
        self.parent.animated_tiles.remove(self.sprite_collection)
        animation_name = self.current_animation_name
        start_time = self.sprite_collection.latest_frame_gt
        self.parent.remove_element(self)
        self.current_animation_name = None  # So new animation doesn't register as the same
        self.sprite_collection = TiledMapSpriteCollection(renpy.loader.transfn(
                "pink_engine/sprite_collections/" + target_sprite_collection), start_time)
        self._set_animation(animation_name)
        self.parent.add_element(self.x, self.y, self.layer, self)

        self.parent.animated_tiles.append(self.sprite_collection)


class TiledMapCamera(object):
    def __init__(self, game_map):
        """
        This class serves as the in-game camera for the Pink Engine maps, with the x and y being used as the
        viewport adjustment variables in Pink Engine map screens.
        :param TiledMapGame game_map: The game map this is the camera of
        """
        config = renpy.store.config
        self.map = game_map

        self.viewport_x_offset = 0
        self.viewport_y_offset = 0
        self.viewport_width = config.screen_width
        self.viewport_height = config.screen_height

        self.x = renpy.display.behavior.Adjustment(value=0, range=game_map.image_size.x)
        self.y = renpy.display.behavior.Adjustment(value=0, range=game_map.image_size.y)

        self.xzoom = renpy.store.pink_tmd_camera_default_zoom  # TODO property of map
        self.yzoom = renpy.store.pink_tmd_camera_default_zoom  # TODO property of map

    def center_on_target(self):
        """
        Makes the camera center on its current target.
        """
        pass


class TiledMapGame(renpy.exports.Displayable):
    tile_class = TiledMapGameTile
    tile_animated_class = TiledMapGameTileAnimated
    sprite_collection_class = TiledMapSpriteCollection

    object_class = TiledMapGameObject
    object_sprite_collection_class = TiledMapGameObjectSpriteCollection

    camera_class = TiledMapCamera

    def __init__(self, map_dict, path, **kwargs):
        """
        The parent class of all Pink Engine game maps.
        :param dict map_dict: The initialization dictionary for this map
        :param str path: The path at which this map can be found.
        """
        renpy.exports.Displayable.__init__(self, **kwargs)  # Py2

        """Location of the file. Also serve as the unique ID for the map."""
        self.path = path

        # ----
        self.last_gt = 0.0

        # ----

        # ---------------------------------Pink Engine Properties-----------------------------------------------------
        self.properties = {}
        if map_dict.get('properties'):
            for layer_property in map_dict['properties']:
                self.properties[layer_property['name']] = layer_property['value']

        # ---------------------------------Simple Global Properties---------------------------------------------------
        self.orientation = map_dict["orientation"]
        self.render_order = map_dict["renderorder"]

        """Size of map in tiles."""
        self.grid_size = Coord(x=map_dict.get('width'), y=map_dict.get('height'))

        """Size of the tiles in the map"""
        self.tile_size = Coord(x=map_dict.get('tilewidth'), y=map_dict.get('tileheight'))

        """Size of map in pixels"""
        if self.orientation == "isometric":
            self.image_size = Coord(
                x=int((self.grid_size.x * self.tile_size.x * 0.5) + (self.grid_size.y * self.tile_size.x * 0.5)),
                y=int((self.grid_size.x * self.tile_size.y * 0.5) + (self.grid_size.y * self.tile_size.y * 0.5)))
        elif self.orientation == "staggered":
            self.image_size = Coord(
                x=int((self.grid_size.x + 0.5) * self.tile_size.x),
                y=int(((self.grid_size.y + 1) * self.tile_size.y * 0.5)))
        elif self.orientation == "hexagonal":
            self.image_size = Coord(
                x=int((self.grid_size.x + 0.5) * self.tile_size.x),
                y=int(((self.grid_size.y + 0.75) * self.tile_size.y * 0.75)))
        else:  # Orthogonal
            self.image_size = Coord(x=self.grid_size.x * self.tile_size.x, y=self.grid_size.y * self.tile_size.y)

        # --- init global
        self.camera = self.camera_class(self)
        self.config = renpy.store.config

        # ---------------------------------Initiate Displayable--------------------------------------------------------
        self._displayable = TiledMap(self, **kwargs)

        # ---
        self.tilesets = self._get_tilesets_from_dict(map_dict)

        # ----
        """A list of all registered images used in this map. Used to start and stop image prediction. Tiles and objects
        add the images they use as they load."""
        self.registered_images = set()
        self.base_tiles = {}
        self.animated_tiles = []
        self.conditional_objects = []

        # ---
        self.layers = []
        for layer_dict in map_dict.get('layers'):
            self.add_layer(layer_dict=layer_dict)

        # ----
        self.event_type_functions = {}
        self.controls_enabled = True

        # ----
        self.start_predicting()

    def move_element(self, element_id, new_x, new_y):
        """
        Moves the element with the given ID to the given new x and y
        :param int element_id: The given element id
        :param int new_x: The new x (in pixels)
        :param int new_y: The new y (in pixels)
        """
        self._displayable.move_element(element_id, new_x, new_y)

    def add_layer(self, layer_dict):
        """
        Adds a new layer to this map and returns it.
        :param dict layer_dict: The initialization dictionary for tne new layer
        :return: the new layer
        :rtype: TiledMapGameLayer
        """
        new_layer = TiledMapGameLayer(self, layer_dict)
        self.layers.append(new_layer)
        return new_layer

    def render(self, width, height, st, at):
        """
        :param int width:
        :param int height:
        :param float st: The shown time (in seconds).
        :param float at: The animation time (in seconds).
        """
        # initiate render object
        render = renpy.exports.Render(width, height)

        if renpy.store.pink_otm_current_map is None:  # prevents crash on completion of game
            return render

        self.game_tick(at)

        renpy.exports.redraw(self, 0)
        t = renpy.display.transform.Transform()
        renpy.exports.render(t, width, height, st, at)  # child render
        return render

    def event(self, ev, x, y, st):  # noqa mandatory variables
        """
        Makes this game object respond to events, such as key presses.
        """
        if self.controls_enabled:
            if ev.type in self.event_type_functions:
                self.event_type_functions[ev.type](self, ev, x, y, st)

    def game_tick(self, gt):
        """
        Core gameplay loop, advancing the game by a single tick per run.
        :param float gt: The current game time in seconds
        """
        self.check_animated_tiles(gt)
        self.check_conditional_objects()
        self.last_gt = gt

    def check_animated_tiles(self, gt):
        """
        Updates the current frame of all animated tiles.
        """
        for animated_tile in self.animated_tiles:
            animated_tile.update_frame(gt)

    def check_conditional_objects(self):
        """
        Checks all conditional objects, adding and removing them according to whether or not their condition is
        currently met.
        """
        for conditional_object in self.conditional_objects:
            self.check_conditional_object(conditional_object)

    def check_conditional_object(self, conditional_object):
        """
        Checks the given object's conditionality, adding and removing it according to whether or not their condition is
        currently met.
        :param TiledMapGameObject conditional_object: The object to check the conditionality of
        """
        if conditional_object.displayable_id is None and conditional_object.condition_met():
            self.add_element(
                conditional_object.x, conditional_object.y, conditional_object.layer, conditional_object)
        elif conditional_object.displayable_id is not None and not conditional_object.condition_met():
            self.remove_element(conditional_object)

    @staticmethod
    def _get_tilesets_from_dict(map_dict):
        tilesets = {}
        for tileset in map_dict.get('tilesets'):
            tileset_source = tileset.get('source')

            """Gets the id at which the tileset starts."""
            first_gid = tileset.get('firstgid')

            """If the tileset has not yet been loaded, does so."""
            tileset_source = renpy.loader.transfn("pink_engine/orthogonal_tiled_maps/" + tileset_source)
            if tileset_source not in pink_tileset_dict:
                Tileset(tileset_source)
            tilesets[first_gid] = pink_tileset_dict[tileset_source]
        return tilesets

    def get_tile_object(self, gid):
        """
        :param int gid: The given gid
        :return: The object from the tileset with the given gid.
        :rtype: TiledMapGameTile
        """
        if gid not in self.base_tiles:
            source_tile = self.get_tile(gid)
            if type(source_tile) is TilesetTile:
                new_tile = self.tile_class(self, source_tile)
            else:
                new_tile = self.tile_animated_class(self, source_tile)  # noqa
                self.animated_tiles.append(new_tile)

            self.base_tiles[gid] = new_tile
        base_tile = self.base_tiles[gid]

        return base_tile

    def add_object(self, object_dict, x, y, layer):
        """
        Adds a new object to this game map
        :param dict object_dict: The object dictionary to initialize the new object
        :param int x: The distance in pixels between the object's image's left side and the left border of the map
        :param int y: The distance in pixels between this object's image's left side and the left border of the map
        :param TiledMapGameLayer layer: The layer on which this object will exist
        """
        # pulled out of the dict)
        property_list = object_dict.get('properties')
        if property_list is None:
            property_list = []

        # Get object properties
        object_properties = {}
        for object_property in property_list:
            object_properties[object_property['name']] = object_property['value']

        # Add base tile properties
        gid = object_dict.get('gid')  # player object doesn't have a gid.
        if gid is not None:
            base_tile = self.get_tile_object(gid)
            for property_key in base_tile.properties:
                if property_key not in object_properties:  # don't override object properties of the same name
                    object_properties[property_key] = base_tile.properties[property_key]

        # Determine if the object is a sprite_collection
        sprite_path = None
        if "sprite_collection_path" in object_properties:
            sprite_path = object_properties['sprite_collection_path']

        # Initialize actual object
        if sprite_path is not None:
            new_object = self.object_sprite_collection_class(x, y, layer, object_properties, sprite_path)
        else:
            width, height, gid = object_dict['width'], object_dict['height'], object_dict['gid']
            base_tile = self.get_tile_object(gid)
            new_object = self.object_class(base_tile, x, y, width, height, layer, object_properties)

        # Add the actual object
        self.add_element(new_object.x, new_object.y, layer, new_object)
        return new_object

    def add_tile(self, tile, x, y, layer):
        """
        Adds an instance of a tile to this game map
        :param self.tile_class|self.tile_animated_class tile: The tile the new tile instance is an instance of
        :param int x: The distance in pixels between the tile's image's left side and the left border of the map
        :param int y: The distance in pixels between this tile's image's left side and the left border of the map
        :param TiledMapGameLayer layer: The layer on which this object will exist
        """
        self.add_element(x, y, layer, tile)

    def add_element(self, x, y, layer, game_object):
        """
        Adds a new element (tile or object) to this game map
        :param int x: The distance in pixels between the tile's image's left side and the left border of the map
        :param int y: The distance in pixels between this tile's image's left side and the left border of the map
        :param TiledMapGameLayer layer: The layer on which this object will exist
        :param self.tile_class|self.tile_animated_class|self.object_class game_object:
        """
        if game_object.is_conditional:
            self.conditional_objects.append(game_object)
            if game_object.condition_met():
                self._displayable.add_element(x=x, y=y, layer=layer, game_object=game_object)
        else:
            self._displayable.add_element(x=x, y=y, layer=layer, game_object=game_object)

    def remove_element(self, element):
        """
        Removes the given element from this map
        :param self.tile_class|self.tile_animated_class|self.object_class element: The element to be removed from this
        map
        """
        element_id = element.displayable_id

        self._displayable.remove_element(element_id)

    def get_displayable(self):
        """
        :return: This object's displayable
        :rtype: TiledMap
        """
        return self._displayable

    def get_tile(self, tile_id):
        """
        Returns the image path for the tile that uses the given ID in this map.

        :param int tile_id: The id used by the given tile on this map. Is equal to the tile's ID within its own tileset
        plus the first_gid for the tileset in this map.
        :return: The tile with the given id.
        :rtype: TilesetTile
        """
        if tile_id == 0:
            return None  # Empty tile

        """First identifies which tileset the image id is from, then retrieves the appropriate tile"""
        first_gids = sorted(self.tilesets, reverse=True)
        prev_gid = first_gids[0]

        while len(first_gids) > 0:
            prev_gid = first_gids.pop(0)
            if tile_id >= prev_gid:
                break

        tile_object = self.tilesets[prev_gid].get_tile(tile_id - prev_gid)

        return tile_object

    def start_predicting(self):
        """
        Causes renpy to start predicting all the images on this map
        """
        for image_name in self.registered_images:
            renpy.exports.start_predict(image_name)

    def stop_predicting(self):
        """
        Causes renpy to stop predicting all the images on this map
        """
        for image_name in self.registered_images:
            renpy.exports.stop_predict(image_name)


class TiledMapGameLayer(object):
    def __init__(self, parent, layer_dict):
        """
        A single layer of a pink engine map
        :param TiledMapGame parent: The game map this layer is a part of
        :param dict layer_dict: The dictionary used to initialize this layer
        """
        self.parent = parent

        # Initializes custom properties
        self.properties = {}
        if layer_dict.get('properties'):
            for layer_property in layer_dict['properties']:
                self.properties[layer_property['name']] = layer_property['value']

        # Executes orientation/type-specific init script.
        if layer_dict.get('type') == "tilelayer" and self.layer_type in LAYER_PRIORITY:
            if self.parent.orientation == "orthogonal":
                self._init_orthogonal_tile_layer(layer_dict=layer_dict)
            elif self.parent.orientation == "isometric":
                self._init_isometric_tile_layer(layer_dict=layer_dict)
            elif self.parent.orientation == "staggered":
                self._init_staggered_tile_layer(layer_dict=layer_dict)
            elif self.parent.orientation == "hexagonal":
                self._init_hexagonal_tile_layer(layer_dict=layer_dict)
        elif layer_dict.get('type') == "objectgroup" and self.layer_type in LAYER_PRIORITY:
            if self.parent.orientation == "orthogonal":
                self._init_orthogonal_object_layer(layer_dict=layer_dict)
            elif self.parent.orientation == "isometric":
                self._init_isometric_object_layer(layer_dict=layer_dict)
            elif self.parent.orientation == "staggered":
                self._init_staggered_object_layer(layer_dict=layer_dict)
            elif self.parent.orientation == "hexagonal":
                self._init_hexagonal_object_layer(layer_dict=layer_dict)

    def _init_orthogonal_tile_layer(self, layer_dict):
        """
        Initializes a tile layer for an orthogonal map
        :param dict layer_dict: The dictionary used to initialize this layer
        """
        xoffset = layer_dict['offsetx'] if layer_dict.get('offsetx') is not None else 0
        yoffset = layer_dict['offsety'] if layer_dict.get('offsety') is not None else 0
        x, c, y = xoffset, 0, yoffset
        if layer_dict.get('data') is None:
            return
        for gid in layer_dict['data']:
            if gid != 0:
                tile = self.parent.get_tile_object(gid)
                place_y = y - tile.height + self.parent.tile_size.y
                self.parent.add_tile(x=x, y=place_y, layer=self, tile=tile)
            x += self.parent.tile_size.x
            c += 1
            if c == self.parent.grid_size.x:
                y += self.parent.tile_size.y
                c, x = 0, xoffset

    def _init_orthogonal_object_layer(self, layer_dict):
        """
        Initializes an object layer for an orthogonal map
        :param dict layer_dict: The dictionary used to initialize this layer
        """
        xoffset = layer_dict['offsetx'] if layer_dict.get('offsetx') is not None else 0
        yoffset = layer_dict['offsety'] if layer_dict.get('offsety') is not None else 0
        if layer_dict.get('objects') is None:
            return
        for layer_object in layer_dict['objects']:
            x = layer_object['x'] + xoffset
            y = layer_object['y'] - layer_object['height'] + yoffset

            self.parent.add_object(layer_object, x, y, self)

    def _init_isometric_tile_layer(self, layer_dict):
        """
        Initializes a tile layer for an isometric map
        :param dict layer_dict: The dictionary used to initialize this layer
        """
        xoffset = layer_dict['offsetx'] if layer_dict.get('offsetx') is not None else 0
        yoffset = layer_dict['offsety'] if layer_dict.get('offsety') is not None else 0
        xcount, ycount = 0, 0
        if layer_dict.get('data') is None:
            return

        for gid in layer_dict['data']:
            if gid != 0:
                tile = self.parent.get_tile_object(gid)
                place_x = ((self.parent.grid_size.y - 1) * self.parent.tile_size.x * 0.5) + \
                          (self.parent.tile_size.x * xcount * 0.5) - (self.parent.tile_size.y * ycount * 0.5) + xoffset
                place_y = (self.parent.tile_size.x * xcount * 0.5) + (self.parent.tile_size.y * ycount * 0.5) - \
                    tile.height + self.parent.tile_size.y + yoffset
                self.parent.add_tile(x=place_x, y=place_y, layer=self, tile=tile)
            xcount += 1
            if xcount == self.parent.grid_size.x:
                xcount = 0
                ycount += 1

    def _init_isometric_object_layer(self, layer_dict):
        """
        Initializes an object layer for an isometric map
        :param dict layer_dict: The dictionary used to initialize this layer
        """
        xoffset = layer_dict['offsetx'] if layer_dict.get('offsetx') is not None else 0
        yoffset = layer_dict['offsety'] if layer_dict.get('offsety') is not None else 0
        if layer_dict.get('objects') is None:
            return
        for layer_object in layer_dict['objects']:
            width, height = layer_object['width'], layer_object['height']
            x = (self.parent.grid_size.y * self.parent.tile_size.x * 0.5) + (layer_object['x'] * 0.5) - \
                (layer_object['y'] * 0.5) - (width * 0.5) + xoffset
            y = (layer_object['x'] * 0.5) + (layer_object['y'] * 0.5) - height + yoffset

            self.parent.add_object(layer_object, x, y, self)

    def _init_staggered_tile_layer(self, layer_dict):
        """
        Initializes a tile layer for a staggered  map
        :param dict layer_dict: The dictionary used to initialize this layer
        """
        xoffset = layer_dict['offsetx'] if layer_dict.get('offsetx') is not None else 0
        yoffset = layer_dict['offsety'] if layer_dict.get('offsety') is not None else 0
        xcount, ycount = 0, 0
        if layer_dict.get('data') is None:
            return
        for gid in layer_dict['data']:
            if gid != 0:
                tile = self.parent.get_tile_object(gid)
                place_x = xcount * self.parent.tile_size.x + xoffset
                if ycount % 2 != 0:
                    place_x += int(self.parent.tile_size.x * 0.5)
                place_y = int(self.parent.tile_size.y * ycount * 0.5) - tile.height + self.parent.tile_size.y + \
                    yoffset
                self.parent.add_tile(x=place_x, y=place_y, layer=self, tile=tile)
            xcount += 1
            if xcount == self.parent.grid_size.x:
                xcount = 0
                ycount += 1

    def _init_staggered_object_layer(self, layer_dict):
        """
        Initializes an object layer for a staggered map
        :param dict layer_dict: The dictionary used to initialize this layer
        """
        xoffset = layer_dict['offsetx'] if layer_dict.get('offsetx') is not None else 0
        yoffset = layer_dict['offsety'] if layer_dict.get('offsety') is not None else 0
        if layer_dict.get('objects') is None:
            return
        for layer_object in layer_dict['objects']:
            x = layer_object['x'] + xoffset
            y = layer_object['y'] - int(layer_object['height']) + yoffset

            self.parent.add_object(layer_object, x, y, self)

    def _init_hexagonal_tile_layer(self, layer_dict):
        """
        Initializes a tile layer for a hexagonal map
        :param dict layer_dict: The dictionary used to initialize this layer
        """
        xoffset = layer_dict['offsetx'] if layer_dict.get('offsetx') is not None else 0
        yoffset = layer_dict['offsety'] if layer_dict.get('offsety') is not None else 0
        xcount, ycount = 0, 0
        if layer_dict.get('data') is None:
            return
        for gid in layer_dict['data']:
            if gid != 0:
                tile = self.parent.get_tile_object(gid)
                place_x = xcount * self.parent.tile_size.x + xoffset
                if ycount % 2 != 0:
                    place_x += int(self.parent.tile_size.x * 0.5)
                place_y = int(self.parent.tile_size.y * ycount * (3.0/4.0)) - tile.height + \
                    self.parent.tile_size.y + yoffset
                self.parent.add_tile(x=place_x, y=place_y, layer=self, tile=tile)
            xcount += 1
            if xcount == self.parent.grid_size.x:
                xcount = 0
                ycount += 1

    def _init_hexagonal_object_layer(self, layer_dict):
        """
        Initializes an object layer for a hexagonal map
        :param dict layer_dict: The dictionary used to initialize this layer
        """
        xoffset = layer_dict['offsetx'] if layer_dict.get('offsetx') is not None else 0
        yoffset = layer_dict['offsety'] if layer_dict.get('offsety') is not None else 0
        if layer_dict.get('objects') is None:
            return
        for layer_object in layer_dict['objects']:
            x = layer_object['x'] + xoffset
            y = layer_object['y'] - int(layer_object['height']) + yoffset + 1

            self.parent.add_object(layer_object, x, y, self)

    @property
    def layer_type(self):
        """
        :return: The layer's type (super, sub, dynamic, movement, interaction). Super renders over dynamic, which
        renders over sub. Independent from the tile layer and object layer distinction.
        :rtype: str
        """
        if "layer_type" in self.properties:
            return self.properties["layer_type"]
        else:
            return "sub"

    @property
    def z_order(self):
        """
        :return: The layer's z-order (priority order for rendering). Lower z-orders of the same type render below higher
        z-orders of the same type.
        :rtype: int
        """
        if "z_order" in self.properties:
            return self.properties["z_order"]
        else:
            return 0


def filepath_to_reg_image(image_filepath):
    """
    :param str image_filepath: The given image filepath
    :return: The
    :rtype: str
    """
    folder_path, file_name = os.path.split(image_filepath)
    file_name, file_ext = os.path.splitext(file_name)
    return renpy.exports.get_registered_image(file_name.lower())


# TODO When checking mobile objects and animated object, use the next gt for their animation/movement, rather than
#  looping through all. That way, objects with slow animations or rare movements won't affect tps.
# TODO restore interaction and movement layer functionality?
